import { CFFParser, CFFCompiler } from './cff_parser.js';
import { SEAC_ANALYSIS_ENABLED, type1FontGlyphMapping } from './fonts_utils.js';
import { warn } from '../shared/util.js';

class CFFFont {
	constructor(file, properties) {
		this.properties = properties;
		const parser = new CFFParser(file, properties, SEAC_ANALYSIS_ENABLED);
		this.cff = parser.parse();
		this.cff.duplicateFirstGlyph();
		const compiler = new CFFCompiler(this.cff);
		this.seacs = this.cff.seacs;

		try {
			this.data = compiler.compile();
		} catch(e) {
			warn("Failed to compile font " + properties.loadedName);
			this.data = file;
		}

		this._createBuiltInEncoding();
	}

	get numGlyphs() {
		return this.cff.charStrings.count;
	}

	getCharset() {
		return this.cff.charset.charset;
	}

	getGlyphMapping() {
		const cff = this.cff;
		const properties = this.properties;
		const { cidToGidMap, cMap } = properties;
		const charsets = cff.charset.charset;
		let charCodeToGlyphId;
		let glyphId;

		if (properties.composite) {
			let invCidToGidMap;

			if (cidToGidMap && cidToGidMap.length > 0) {
				invCidToGidMap = Object.create(null);

				for (let i = 0, ii = cidToGidMap.length; i < ii; i++) {
					const gid = cidToGidMap[i];

					if (gid !== undefined) {
						invCidToGidMap[gid] = i;
					}
				}
			}

			charCodeToGlyphId = Object.create(null);
			let charCode;

			if (cff.isCIDFont) {
				for (glyphId = 0; glyphId < charsets.length; glyphId++) {
					const cid = charsets[glyphId];
					charCode = cMap.charCodeOf(cid);

					if (invCidToGidMap && invCidToGidMap[charCode] !== undefined) {
						charCode = invCidToGidMap[charCode];
					}

					charCodeToGlyphId[charCode] = glyphId;
				}
			} else {
				for (glyphId = 0; glyphId < cff.charStrings.count; glyphId++) {
					charCode = cMap.charCodeOf(glyphId);
					charCodeToGlyphId[charCode] = glyphId;
				}
			}

			return charCodeToGlyphId;
		}

		let encoding = cff.encoding ? cff.encoding.encoding : null;

		if (properties.isInternalFont) {
			encoding = properties.defaultEncoding;
		}

		charCodeToGlyphId = type1FontGlyphMapping(properties, encoding, charsets);
		return charCodeToGlyphId;
	}

	hasGlyphId(id) {
		return this.cff.hasGlyphId(id);
	}

	_createBuiltInEncoding() {
		const { charset, encoding } = this.cff;

		if (!charset || !encoding) {
			return;
		}

		const charsets = charset.charset, encodings = encoding.encoding;
		const map = [];

		for (const charCode in encodings) {
			const glyphId = encodings[charCode];

			if (glyphId >= 0) {
				const glyphName = charsets[glyphId];

				if (glyphName) {
					map[charCode] = glyphName;
				}
			}
		}

		if (map.length > 0) {
			this.properties.builtInEncoding = map;
		}
	}
}

export { CFFFont };
