import { getEncoding, StandardEncoding } from './encodings.js';
import { getGlyphsUnicode } from './glyphlist.js';
import { getUnicodeForGlyph } from './unicode.js';
import { info } from '../shared/util.js';

const SEAC_ANALYSIS_ENABLED = true;

const FontFlags = {
	FixedPitch: 1,
	Serif: 2,
	Symbolic: 4,
	Script: 8,
	Nonsymbolic: 32,
	Italic: 64,
	AllCap: 65536,
	SmallCap: 131072,
	ForceBold: 262144
};

const MacStandardGlyphOrdering = [
	".notdef",
	".null",
	"nonmarkingreturn",
	"space",
	"exclam",
	"quotedbl",
	"numbersign",
	"dollar",
	"percent",
	"ampersand",
	"quotesingle",
	"parenleft",
	"parenright",
	"asterisk",
	"plus",
	"comma",
	"hyphen",
	"period",
	"slash",
	"zero",
	"one",
	"two",
	"three",
	"four",
	"five",
	"six",
	"seven",
	"eight",
	"nine",
	"colon",
	"semicolon",
	"less",
	"equal",
	"greater",
	"question",
	"at",
	"A",
	"B",
	"C",
	"D",
	"E",
	"F",
	"G",
	"H",
	"I",
	"J",
	"K",
	"L",
	"M",
	"N",
	"O",
	"P",
	"Q",
	"R",
	"S",
	"T",
	"U",
	"V",
	"W",
	"X",
	"Y",
	"Z",
	"bracketleft",
	"backslash",
	"bracketright",
	"asciicircum",
	"underscore",
	"grave",
	"a",
	"b",
	"c",
	"d",
	"e",
	"f",
	"g",
	"h",
	"i",
	"j",
	"k",
	"l",
	"m",
	"n",
	"o",
	"p",
	"q",
	"r",
	"s",
	"t",
	"u",
	"v",
	"w",
	"x",
	"y",
	"z",
	"braceleft",
	"bar",
	"braceright",
	"asciitilde",
	"Adieresis",
	"Aring",
	"Ccedilla",
	"Eacute",
	"Ntilde",
	"Odieresis",
	"Udieresis",
	"aacute",
	"agrave",
	"acircumflex",
	"adieresis",
	"atilde",
	"aring",
	"ccedilla",
	"eacute",
	"egrave",
	"ecircumflex",
	"edieresis",
	"iacute",
	"igrave",
	"icircumflex",
	"idieresis",
	"ntilde",
	"oacute",
	"ograve",
	"ocircumflex",
	"odieresis",
	"otilde",
	"uacute",
	"ugrave",
	"ucircumflex",
	"udieresis",
	"dagger",
	"degree",
	"cent",
	"sterling",
	"section",
	"bullet",
	"paragraph",
	"germandbls",
	"registered",
	"copyright",
	"trademark",
	"acute",
	"dieresis",
	"notequal",
	"AE",
	"Oslash",
	"infinity",
	"plusminus",
	"lessequal",
	"greaterequal",
	"yen",
	"mu",
	"partialdiff",
	"summation",
	"product",
	"pi",
	"integral",
	"ordfeminine",
	"ordmasculine",
	"Omega",
	"ae",
	"oslash",
	"questiondown",
	"exclamdown",
	"logicalnot",
	"radical",
	"florin",
	"approxequal",
	"Delta",
	"guillemotleft",
	"guillemotright",
	"ellipsis",
	"nonbreakingspace",
	"Agrave",
	"Atilde",
	"Otilde",
	"OE",
	"oe",
	"endash",
	"emdash",
	"quotedblleft",
	"quotedblright",
	"quoteleft",
	"quoteright",
	"divide",
	"lozenge",
	"ydieresis",
	"Ydieresis",
	"fraction",
	"currency",
	"guilsinglleft",
	"guilsinglright",
	"fi",
	"fl",
	"daggerdbl",
	"periodcentered",
	"quotesinglbase",
	"quotedblbase",
	"perthousand",
	"Acircumflex",
	"Ecircumflex",
	"Aacute",
	"Edieresis",
	"Egrave",
	"Iacute",
	"Icircumflex",
	"Idieresis",
	"Igrave",
	"Oacute",
	"Ocircumflex",
	"apple",
	"Ograve",
	"Uacute",
	"Ucircumflex",
	"Ugrave",
	"dotlessi",
	"circumflex",
	"tilde",
	"macron",
	"breve",
	"dotaccent",
	"ring",
	"cedilla",
	"hungarumlaut",
	"ogonek",
	"caron",
	"Lslash",
	"lslash",
	"Scaron",
	"scaron",
	"Zcaron",
	"zcaron",
	"brokenbar",
	"Eth",
	"eth",
	"Yacute",
	"yacute",
	"Thorn",
	"thorn",
	"minus",
	"multiply",
	"onesuperior",
	"twosuperior",
	"threesuperior",
	"onehalf",
	"onequarter",
	"threequarters",
	"franc",
	"Gbreve",
	"gbreve",
	"Idotaccent",
	"Scedilla",
	"scedilla",
	"Cacute",
	"cacute",
	"Ccaron",
	"ccaron",
	"dcroat"
];

function recoverGlyphName(name, glyphsUnicodeMap) {
	if (glyphsUnicodeMap[name] !== undefined) {
		return name;
	}

	const unicode = getUnicodeForGlyph(name, glyphsUnicodeMap);

	if (unicode !== -1) {
		for (const key in glyphsUnicodeMap) {
			if (glyphsUnicodeMap[key] === unicode) {
				return key;
			}
		}
	}

	info("Unable to recover a standard glyph name for: " + name);
	return name;
}

function type1FontGlyphMapping(properties, builtInEncoding, glyphNames) {
	const charCodeToGlyphId = Object.create(null);
	let glyphId, charCode, baseEncoding;
	const isSymbolicFont = !!(properties.flags & FontFlags.Symbolic);

	if (properties.isInternalFont) {
		baseEncoding = builtInEncoding;

		for (charCode = 0; charCode < baseEncoding.length; charCode++) {
			glyphId = glyphNames.indexOf(baseEncoding[charCode]);

			if (glyphId >= 0) {
				charCodeToGlyphId[charCode] = glyphId;
			} else {
				charCodeToGlyphId[charCode] = 0;
			}
		}
	} else if (properties.baseEncodingName) {
		baseEncoding = getEncoding(properties.baseEncodingName);

		for (charCode = 0; charCode < baseEncoding.length; charCode++) {
			glyphId = glyphNames.indexOf(baseEncoding[charCode]);

			if (glyphId >= 0) {
				charCodeToGlyphId[charCode] = glyphId;
			} else {
				charCodeToGlyphId[charCode] = 0;
			}
		}
	} else if (isSymbolicFont) {
		for (charCode in builtInEncoding) {
			charCodeToGlyphId[charCode] = builtInEncoding[charCode];
		}
	} else {
		baseEncoding = StandardEncoding;

		for (charCode = 0; charCode < baseEncoding.length; charCode++) {
			glyphId = glyphNames.indexOf(baseEncoding[charCode]);

			if (glyphId >= 0) {
				charCodeToGlyphId[charCode] = glyphId;
			} else {
				charCodeToGlyphId[charCode] = 0;
			}
		}
	}

	const differences = properties.differences;
	let glyphsUnicodeMap;

	if (differences) {
		for (charCode in differences) {
			const glyphName = differences[charCode];
			glyphId = glyphNames.indexOf(glyphName);

			if (glyphId === -1) {
				if (!glyphsUnicodeMap) {
					glyphsUnicodeMap = getGlyphsUnicode();
				}

				const standardGlyphName = recoverGlyphName(glyphName, glyphsUnicodeMap);

				if (standardGlyphName !== glyphName) {
					glyphId = glyphNames.indexOf(standardGlyphName);
				}
			}

			if (glyphId >= 0) {
				charCodeToGlyphId[charCode] = glyphId;
			} else {
				charCodeToGlyphId[charCode] = 0;
			}
		}
	}

	return charCodeToGlyphId;
}

function normalizeFontName(name) {
	return name.replaceAll(/[,_]/g, "-").replaceAll(/\s/g, "");
}

export { FontFlags, MacStandardGlyphOrdering, SEAC_ANALYSIS_ENABLED, normalizeFontName, recoverGlyphName, type1FontGlyphMapping };
