import { DecodeStream } from './decode_stream.js';
import { Dict } from './primitives.js';
import { JpegImage } from './jpg.js';
import { shadow } from '../shared/util.js';

class JpegStream extends DecodeStream {
	constructor(stream, maybeLength, params) {
		let ch;

		while ((ch = stream.getByte()) !== -1) {
			if (ch === 0xff) {
				stream.skip(-1);
				break;
			}
		}

		super(maybeLength);
		this.stream = stream;
		this.dict = stream.dict;
		this.maybeLength = maybeLength;
		this.params = params;
	}

	get bytes() {
		return shadow(this, "bytes", this.stream.getBytes(this.maybeLength));
	}

	ensureBuffer(requested) {
		
	}

	readBlock() {
		if (this.eof) {
			return;
		}

		const jpegOptions = {
			decodeTransform: undefined,
			colorTransform: undefined
		};

		const decodeArr = this.dict.getArray("D", "Decode");

		if ((this.forceRGBA || this.forceRGB) && Array.isArray(decodeArr)) {
			const bitsPerComponent = this.dict.get("BPC", "BitsPerComponent") || 8;
			const decodeArrLength = decodeArr.length;
			const transform = new Int32Array(decodeArrLength);
			let transformNeeded = false;
			const maxValue = (1 << bitsPerComponent) - 1;

			for (let i = 0; i < decodeArrLength; i += 2) {
				transform[i] = (decodeArr[i + 1] - decodeArr[i]) * 256 | 0;
				transform[i + 1] = decodeArr[i] * maxValue | 0;

				if (transform[i] !== 256 || transform[i + 1] !== 0) {
					transformNeeded = true;
				}
			}

			if (transformNeeded) {
				jpegOptions.decodeTransform = transform;
			}
		}

		if (this.params instanceof Dict) {
			const colorTransform = this.params.get("ColorTransform");

			if (Number.isInteger(colorTransform)) {
				jpegOptions.colorTransform = colorTransform;
			}
		}

		const jpegImage = new JpegImage(jpegOptions);
		jpegImage.parse(this.bytes);

		const data = jpegImage.getData({
			width: this.drawWidth,
			height: this.drawHeight,
			forceRGBA: this.forceRGBA,
			forceRGB: this.forceRGB,
			isSourcePDF: true
		});

		this.buffer = data;
		this.bufferLength = data.length;
		this.eof = true;
	}
}

export { JpegStream };
