import { unreachable, shadow } from '../shared/util.js';

const CIRCULAR_REF = Symbol("CIRCULAR_REF");
const EOF = Symbol("EOF");
let CmdCache = Object.create(null);
let NameCache = Object.create(null);
let RefCache = Object.create(null);

function clearPrimitiveCaches() {
	CmdCache = Object.create(null);
	NameCache = Object.create(null);
	RefCache = Object.create(null);
}

class Name {
	constructor(name) {
		this.name = name;
	}

	static get(name) {
		return NameCache[name] || (NameCache[name] = new Name(name));
	}
}

class Cmd {
	constructor(cmd) {
		this.cmd = cmd;
	}

	static get(cmd) {
		return CmdCache[cmd] || (CmdCache[cmd] = new Cmd(cmd));
	}
}

const nonSerializable = function nonSerializableClosure() {
	return nonSerializable;
};

class Dict {
	constructor(xref = null) {
		this._map = Object.create(null);
		this.xref = xref;
		this.objId = null;
		this.suppressEncryption = false;
		this.__nonSerializable__ = nonSerializable;
	}

	assignXref(newXref) {
		this.xref = newXref;
	}

	get size() {
		return Object.keys(this._map).length;
	}

	get(key1, key2, key3) {
		let value = this._map[key1];

		if (value === undefined && key2 !== undefined) {
			value = this._map[key2];

			if (value === undefined && key3 !== undefined) {
				value = this._map[key3];
			}
		}

		if (value instanceof Ref && this.xref) {
			return this.xref.fetch(value, this.suppressEncryption);
		}

		return value;
	}

	async getAsync(key1, key2, key3) {
		let value = this._map[key1];

		if (value === undefined && key2 !== undefined) {
			value = this._map[key2];

			if (value === undefined && key3 !== undefined) {
				value = this._map[key3];
			}
		}

		if (value instanceof Ref && this.xref) {
			return this.xref.fetchAsync(value, this.suppressEncryption);
		}

		return value;
	}

	getArray(key1, key2, key3) {
		let value = this._map[key1];

		if (value === undefined && key2 !== undefined) {
			value = this._map[key2];

			if (value === undefined && key3 !== undefined) {
				value = this._map[key3];
			}
		}

		if (value instanceof Ref && this.xref) {
			value = this.xref.fetch(value, this.suppressEncryption);
		}

		if (Array.isArray(value)) {
			value = value.slice();

			for (let i = 0, ii = value.length; i < ii; i++) {
				if (value[i] instanceof Ref && this.xref) {
					value[i] = this.xref.fetch(value[i], this.suppressEncryption);
				}
			}
		}

		return value;
	}

	getRaw(key) {
		return this._map[key];
	}

	getKeys() {
		return Object.keys(this._map);
	}

	getRawValues() {
		return Object.values(this._map);
	}

	set(key, value) {
		this._map[key] = value;
	}

	has(key) {
		return this._map[key] !== undefined;
	}

	forEach(callback) {
		for (const key in this._map) {
			callback(key, this.get(key));
		}
	}

	static get empty() {
		const emptyDict = new Dict(null);

		emptyDict.set = (key, value) => {
			unreachable("Should not call `set` on the empty dictionary.");
		};

		return shadow(this, "empty", emptyDict);
	}

	static merge({ xref, dictArray, mergeSubDicts = false }) {
		const mergedDict = new Dict(xref), properties = new Map();

		for (const dict of dictArray) {
			if (!(dict instanceof Dict)) {
				continue;
			}

			for (const [key, value] of Object.entries(dict._map)) {
				let property = properties.get(key);

				if (property === undefined) {
					property = [];
					properties.set(key, property);
				} else if (!mergeSubDicts || !(value instanceof Dict)) {
					continue;
				}

				property.push(value);
			}
		}

		for (const [name, values] of properties) {
			if (values.length === 1 || !(values[0] instanceof Dict)) {
				mergedDict._map[name] = values[0];
				continue;
			}

			const subDict = new Dict(xref);

			for (const dict of values) {
				for (const [key, value] of Object.entries(dict._map)) {
					if (subDict._map[key] === undefined) {
						subDict._map[key] = value;
					}
				}
			}

			if (subDict.size > 0) {
				mergedDict._map[name] = subDict;
			}
		}

		properties.clear();
		return mergedDict.size > 0 ? mergedDict : Dict.empty;
	}
}

class Ref {
	constructor(num, gen) {
		this.num = num;
		this.gen = gen;
	}

	toString() {
		if (this.gen === 0) {
			return `${this.num}R`;
		}

		return `${this.num}R${this.gen}`;
	}

	static get(num, gen) {
		const key = gen === 0 ? `${num}R` : `${num}R${gen}`;
		return RefCache[key] || (RefCache[key] = new Ref(num, gen));
	}
}

class RefSet {
	constructor(parent = null) {
		this._set = new Set(parent && parent._set);
	}

	has(ref) {
		return this._set.has(ref.toString());
	}

	put(ref) {
		this._set.add(ref.toString());
	}

	remove(ref) {
		this._set.delete(ref.toString());
	}

	[Symbol.iterator]() {
		return this._set.values();
	}

	clear() {
		this._set.clear();
	}
}

class RefSetCache {
	constructor() {
		this._map = new Map();
	}

	get size() {
		return this._map.size;
	}

	get(ref) {
		return this._map.get(ref.toString());
	}

	has(ref) {
		return this._map.has(ref.toString());
	}

	put(ref, obj) {
		this._map.set(ref.toString(), obj);
	}

	putAlias(ref, aliasRef) {
		this._map.set(ref.toString(), this.get(aliasRef));
	}

	[Symbol.iterator]() {
		return this._map.values();
	}

	clear() {
		this._map.clear();
	}
}

function isName(v, name) {
	return v instanceof Name && (name === undefined || v.name === name);
}

function isCmd(v, cmd) {
	return v instanceof Cmd && (cmd === undefined || v.cmd === cmd);
}

function isDict(v, type) {
	return v instanceof Dict && (type === undefined || isName(v.get("Type"), type));
}

function isRefsEqual(v1, v2) {
	return v1.num === v2.num && v1.gen === v2.gen;
}

export { CIRCULAR_REF, Cmd, Dict, EOF, Name, Ref, RefSet, RefSetCache, clearPrimitiveCaches, isCmd, isDict, isName, isRefsEqual };
