import { NamespaceIds, $buildXFAObject } from './namespaces.js';
import { XFAObject, XFAObjectArray, StringObject, ContentObject } from './xfa_object.js';
import { getStringOption, getInteger } from './utils.js';

const LOCALE_SET_NS_ID = NamespaceIds.localeSet.id;

class CalendarSymbols extends XFAObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "calendarSymbols", true);
		this.name = "gregorian";
		this.dayNames = new XFAObjectArray(2);
		this.eraNames = null;
		this.meridiemNames = null;
		this.monthNames = new XFAObjectArray(2);
	}
}

class CurrencySymbol extends StringObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "currencySymbol");
		this.name = getStringOption(attributes.name, ["symbol", "isoname", "decimal"]);
	}
}

class CurrencySymbols extends XFAObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "currencySymbols", true);
		this.currencySymbol = new XFAObjectArray(3);
	}
}

class DatePattern extends StringObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "datePattern");
		this.name = getStringOption(attributes.name, ["full", "long", "med", "short"]);
	}
}

class DatePatterns extends XFAObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "datePatterns", true);
		this.datePattern = new XFAObjectArray(4);
	}
}

class DateTimeSymbols extends ContentObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "dateTimeSymbols");
	}
}

class Day extends StringObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "day");
	}
}

class DayNames extends XFAObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "dayNames", true);

		this.abbr = getInteger({
			data: attributes.abbr,
			defaultValue: 0,
			validate: x => x === 1
		});

		this.day = new XFAObjectArray(7);
	}
}

class Era extends StringObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "era");
	}
}

class EraNames extends XFAObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "eraNames", true);
		this.era = new XFAObjectArray(2);
	}
}

class Locale extends XFAObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "locale", true);
		this.desc = attributes.desc || "";
		this.name = "isoname";
		this.calendarSymbols = null;
		this.currencySymbols = null;
		this.datePatterns = null;
		this.dateTimeSymbols = null;
		this.numberPatterns = null;
		this.numberSymbols = null;
		this.timePatterns = null;
		this.typeFaces = null;
	}
}

class LocaleSet extends XFAObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "localeSet", true);
		this.locale = new XFAObjectArray();
	}
}

class Meridiem extends StringObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "meridiem");
	}
}

class MeridiemNames extends XFAObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "meridiemNames", true);
		this.meridiem = new XFAObjectArray(2);
	}
}

class Month extends StringObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "month");
	}
}

class MonthNames extends XFAObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "monthNames", true);

		this.abbr = getInteger({
			data: attributes.abbr,
			defaultValue: 0,
			validate: x => x === 1
		});

		this.month = new XFAObjectArray(12);
	}
}

class NumberPattern extends StringObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "numberPattern");
		this.name = getStringOption(attributes.name, ["full", "long", "med", "short"]);
	}
}

class NumberPatterns extends XFAObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "numberPatterns", true);
		this.numberPattern = new XFAObjectArray(4);
	}
}

class NumberSymbol extends StringObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "numberSymbol");
		this.name = getStringOption(attributes.name, ["decimal", "grouping", "percent", "minus", "zero"]);
	}
}

class NumberSymbols extends XFAObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "numberSymbols", true);
		this.numberSymbol = new XFAObjectArray(5);
	}
}

class TimePattern extends StringObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "timePattern");
		this.name = getStringOption(attributes.name, ["full", "long", "med", "short"]);
	}
}

class TimePatterns extends XFAObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "timePatterns", true);
		this.timePattern = new XFAObjectArray(4);
	}
}

class TypeFace extends XFAObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "typeFace", true);
		this.name = attributes.name | "";
	}
}

class TypeFaces extends XFAObject {
	constructor(attributes) {
		super(LOCALE_SET_NS_ID, "typeFaces", true);
		this.typeFace = new XFAObjectArray();
	}
}

class LocaleSetNamespace {
	static [$buildXFAObject](name, attributes) {
		if (LocaleSetNamespace.hasOwnProperty(name)) {
			return LocaleSetNamespace[name](attributes);
		}

		return undefined;
	}

	static calendarSymbols(attrs) {
		return new CalendarSymbols(attrs);
	}

	static currencySymbol(attrs) {
		return new CurrencySymbol(attrs);
	}

	static currencySymbols(attrs) {
		return new CurrencySymbols(attrs);
	}

	static datePattern(attrs) {
		return new DatePattern(attrs);
	}

	static datePatterns(attrs) {
		return new DatePatterns(attrs);
	}

	static dateTimeSymbols(attrs) {
		return new DateTimeSymbols(attrs);
	}

	static day(attrs) {
		return new Day(attrs);
	}

	static dayNames(attrs) {
		return new DayNames(attrs);
	}

	static era(attrs) {
		return new Era(attrs);
	}

	static eraNames(attrs) {
		return new EraNames(attrs);
	}

	static locale(attrs) {
		return new Locale(attrs);
	}

	static localeSet(attrs) {
		return new LocaleSet(attrs);
	}

	static meridiem(attrs) {
		return new Meridiem(attrs);
	}

	static meridiemNames(attrs) {
		return new MeridiemNames(attrs);
	}

	static month(attrs) {
		return new Month(attrs);
	}

	static monthNames(attrs) {
		return new MonthNames(attrs);
	}

	static numberPattern(attrs) {
		return new NumberPattern(attrs);
	}

	static numberPatterns(attrs) {
		return new NumberPatterns(attrs);
	}

	static numberSymbol(attrs) {
		return new NumberSymbol(attrs);
	}

	static numberSymbols(attrs) {
		return new NumberSymbols(attrs);
	}

	static timePattern(attrs) {
		return new TimePattern(attrs);
	}

	static timePatterns(attrs) {
		return new TimePatterns(attrs);
	}

	static typeFace(attrs) {
		return new TypeFace(attrs);
	}

	static typeFaces(attrs) {
		return new TypeFaces(attrs);
	}
}

export { LocaleSetNamespace };
