import { CalibriRegularFactors, CalibriRegularMetrics, CalibriBoldFactors, CalibriBoldMetrics, CalibriItalicFactors, CalibriItalicMetrics, CalibriBoldItalicFactors, CalibriBoldItalicMetrics } from './calibri_factors.js';
import { Name, Dict } from './primitives.js';
import { HelveticaRegularFactors, HelveticaRegularMetrics, HelveticaBoldFactors, HelveticaBoldMetrics, HelveticaItalicFactors, HelveticaItalicMetrics, HelveticaBoldItalicFactors, HelveticaBoldItalicMetrics } from './helvetica_factors.js';
import { LiberationSansRegularWidths, LiberationSansRegularMapping, LiberationSansBoldWidths, LiberationSansBoldMapping, LiberationSansItalicWidths, LiberationSansItalicMapping, LiberationSansBoldItalicWidths, LiberationSansBoldItalicMapping } from './liberationsans_widths.js';
import { MyriadProRegularFactors, MyriadProRegularMetrics, MyriadProBoldFactors, MyriadProBoldMetrics, MyriadProItalicFactors, MyriadProItalicMetrics, MyriadProBoldItalicFactors, MyriadProBoldItalicMetrics } from './myriadpro_factors.js';
import { SegoeuiRegularFactors, SegoeuiRegularMetrics, SegoeuiBoldFactors, SegoeuiBoldMetrics, SegoeuiItalicFactors, SegoeuiItalicMetrics, SegoeuiBoldItalicFactors, SegoeuiBoldItalicMetrics } from './segoeui_factors.js';
import { getLookupTableFactory } from './core_utils.js';
import { normalizeFontName } from './fonts_utils.js';

const getXFAFontMap = getLookupTableFactory(function (t) {
	t["MyriadPro-Regular"] = t["PdfJS-Fallback-Regular"] = {
		name: "LiberationSans-Regular",
		factors: MyriadProRegularFactors,
		baseWidths: LiberationSansRegularWidths,
		baseMapping: LiberationSansRegularMapping,
		metrics: MyriadProRegularMetrics
	};

	t["MyriadPro-Bold"] = t["PdfJS-Fallback-Bold"] = {
		name: "LiberationSans-Bold",
		factors: MyriadProBoldFactors,
		baseWidths: LiberationSansBoldWidths,
		baseMapping: LiberationSansBoldMapping,
		metrics: MyriadProBoldMetrics
	};

	t["MyriadPro-It"] = t["MyriadPro-Italic"] = t["PdfJS-Fallback-Italic"] = {
		name: "LiberationSans-Italic",
		factors: MyriadProItalicFactors,
		baseWidths: LiberationSansItalicWidths,
		baseMapping: LiberationSansItalicMapping,
		metrics: MyriadProItalicMetrics
	};

	t["MyriadPro-BoldIt"] = t["MyriadPro-BoldItalic"] = t["PdfJS-Fallback-BoldItalic"] = {
		name: "LiberationSans-BoldItalic",
		factors: MyriadProBoldItalicFactors,
		baseWidths: LiberationSansBoldItalicWidths,
		baseMapping: LiberationSansBoldItalicMapping,
		metrics: MyriadProBoldItalicMetrics
	};

	t.ArialMT = t.Arial = t["Arial-Regular"] = {
		name: "LiberationSans-Regular",
		baseWidths: LiberationSansRegularWidths,
		baseMapping: LiberationSansRegularMapping
	};

	t["Arial-BoldMT"] = t["Arial-Bold"] = {
		name: "LiberationSans-Bold",
		baseWidths: LiberationSansBoldWidths,
		baseMapping: LiberationSansBoldMapping
	};

	t["Arial-ItalicMT"] = t["Arial-Italic"] = {
		name: "LiberationSans-Italic",
		baseWidths: LiberationSansItalicWidths,
		baseMapping: LiberationSansItalicMapping
	};

	t["Arial-BoldItalicMT"] = t["Arial-BoldItalic"] = {
		name: "LiberationSans-BoldItalic",
		baseWidths: LiberationSansBoldItalicWidths,
		baseMapping: LiberationSansBoldItalicMapping
	};

	t["Calibri-Regular"] = {
		name: "LiberationSans-Regular",
		factors: CalibriRegularFactors,
		baseWidths: LiberationSansRegularWidths,
		baseMapping: LiberationSansRegularMapping,
		metrics: CalibriRegularMetrics
	};

	t["Calibri-Bold"] = {
		name: "LiberationSans-Bold",
		factors: CalibriBoldFactors,
		baseWidths: LiberationSansBoldWidths,
		baseMapping: LiberationSansBoldMapping,
		metrics: CalibriBoldMetrics
	};

	t["Calibri-Italic"] = {
		name: "LiberationSans-Italic",
		factors: CalibriItalicFactors,
		baseWidths: LiberationSansItalicWidths,
		baseMapping: LiberationSansItalicMapping,
		metrics: CalibriItalicMetrics
	};

	t["Calibri-BoldItalic"] = {
		name: "LiberationSans-BoldItalic",
		factors: CalibriBoldItalicFactors,
		baseWidths: LiberationSansBoldItalicWidths,
		baseMapping: LiberationSansBoldItalicMapping,
		metrics: CalibriBoldItalicMetrics
	};

	t["Segoeui-Regular"] = {
		name: "LiberationSans-Regular",
		factors: SegoeuiRegularFactors,
		baseWidths: LiberationSansRegularWidths,
		baseMapping: LiberationSansRegularMapping,
		metrics: SegoeuiRegularMetrics
	};

	t["Segoeui-Bold"] = {
		name: "LiberationSans-Bold",
		factors: SegoeuiBoldFactors,
		baseWidths: LiberationSansBoldWidths,
		baseMapping: LiberationSansBoldMapping,
		metrics: SegoeuiBoldMetrics
	};

	t["Segoeui-Italic"] = {
		name: "LiberationSans-Italic",
		factors: SegoeuiItalicFactors,
		baseWidths: LiberationSansItalicWidths,
		baseMapping: LiberationSansItalicMapping,
		metrics: SegoeuiItalicMetrics
	};

	t["Segoeui-BoldItalic"] = {
		name: "LiberationSans-BoldItalic",
		factors: SegoeuiBoldItalicFactors,
		baseWidths: LiberationSansBoldItalicWidths,
		baseMapping: LiberationSansBoldItalicMapping,
		metrics: SegoeuiBoldItalicMetrics
	};

	t["Helvetica-Regular"] = t.Helvetica = {
		name: "LiberationSans-Regular",
		factors: HelveticaRegularFactors,
		baseWidths: LiberationSansRegularWidths,
		baseMapping: LiberationSansRegularMapping,
		metrics: HelveticaRegularMetrics
	};

	t["Helvetica-Bold"] = {
		name: "LiberationSans-Bold",
		factors: HelveticaBoldFactors,
		baseWidths: LiberationSansBoldWidths,
		baseMapping: LiberationSansBoldMapping,
		metrics: HelveticaBoldMetrics
	};

	t["Helvetica-Italic"] = {
		name: "LiberationSans-Italic",
		factors: HelveticaItalicFactors,
		baseWidths: LiberationSansItalicWidths,
		baseMapping: LiberationSansItalicMapping,
		metrics: HelveticaItalicMetrics
	};

	t["Helvetica-BoldItalic"] = {
		name: "LiberationSans-BoldItalic",
		factors: HelveticaBoldItalicFactors,
		baseWidths: LiberationSansBoldItalicWidths,
		baseMapping: LiberationSansBoldItalicMapping,
		metrics: HelveticaBoldItalicMetrics
	};
});

function getXfaFontName(name) {
	const fontName = normalizeFontName(name);
	const fontMap = getXFAFontMap();
	return fontMap[fontName];
}

function getXfaFontWidths(name) {
	const info = getXfaFontName(name);

	if (!info) {
		return null;
	}

	const { baseWidths, baseMapping, factors } = info;
	let rescaledBaseWidths;

	if (!factors) {
		rescaledBaseWidths = baseWidths;
	} else {
		rescaledBaseWidths = baseWidths.map((w, i) => w * factors[i]);
	}

	let currentCode = -2;
	let currentArray;
	const newWidths = [];

	for (const [unicode, glyphIndex] of baseMapping.map((charUnicode, index) => [charUnicode, index]).sort(([unicode1], [unicode2]) => unicode1 - unicode2)) {
		if (unicode === -1) {
			continue;
		}

		if (unicode === currentCode + 1) {
			currentArray.push(rescaledBaseWidths[glyphIndex]);
			currentCode += 1;
		} else {
			currentCode = unicode;
			currentArray = [rescaledBaseWidths[glyphIndex]];
			newWidths.push(unicode, currentArray);
		}
	}

	return newWidths;
}

function getXfaFontDict(name) {
	const widths = getXfaFontWidths(name);
	const dict = new Dict(null);
	dict.set("BaseFont", Name.get(name));
	dict.set("Type", Name.get("Font"));
	dict.set("Subtype", Name.get("CIDFontType2"));
	dict.set("Encoding", Name.get("Identity-H"));
	dict.set("CIDToGIDMap", Name.get("Identity"));
	dict.set("W", widths);
	dict.set("FirstChar", widths[0]);
	dict.set("LastChar", widths.at(-2) + widths.at(-1).length - 1);
	const descriptor = new Dict(null);
	dict.set("FontDescriptor", descriptor);
	const systemInfo = new Dict(null);
	systemInfo.set("Ordering", "Identity");
	systemInfo.set("Registry", "Adobe");
	systemInfo.set("Supplement", 0);
	dict.set("CIDSystemInfo", systemInfo);
	return dict;
}

export { getXfaFontDict, getXfaFontName };
