import require$$0 from '../../_virtual/node_utils.js';
import require$$1 from '../../_virtual/node_stream.js';
import require$$2 from '../../_virtual/network.js';
import require$$3 from '../../_virtual/fetch_stream.js';
import { isArrayBuffer, setVerbosityLevel, stringToBytes, createPromiseCapability, unreachable, getVerbosityLevel, info, warn, shadow, AnnotationMode, RenderingIntentFlag, AbortException, assert, UnknownErrorException, UnexpectedResponseException, MissingPDFException, InvalidPDFException, PasswordException, MAX_IMAGE_SIZE_TO_CACHE } from '../shared/util.js';
import { AnnotationStorage, PrintAnnotationStorage } from './annotation_storage.js';
import { isDataScheme, DOMCMapReaderFactory, DOMStandardFontDataFactory, isValidFetchUrl, deprecated, DOMCanvasFactory, DOMFilterFactory, StatTimer, PageViewport, RenderingCancelledException, loadScript } from './display_utils.js';
import { FontLoader, FontFaceObject } from './font_loader.js';
import { CanvasGraphics } from './canvas.js';
import { GlobalWorkerOptions } from './worker_options.js';
import { isNodeJS } from '../shared/is_node.js';
import { MessageHandler } from '../shared/message_handler.js';
import { Metadata } from './metadata.js';
import { OptionalContentConfig } from './optional_content_config.js';
import { PDFDataTransportStream } from './transport_stream.js';
import { XfaText } from './xfa_text.js';

const DEFAULT_RANGE_CHUNK_SIZE = 65536;
const RENDERING_CANCELLED_TIMEOUT = 100;
const DELAYED_CLEANUP_TIMEOUT = 5000;
let DefaultCanvasFactory = DOMCanvasFactory;
let DefaultCMapReaderFactory = DOMCMapReaderFactory;
let DefaultFilterFactory = DOMFilterFactory;
let DefaultStandardFontDataFactory = DOMStandardFontDataFactory;

if (isNodeJS) {
	const { NodeCanvasFactory, NodeCMapReaderFactory, NodeFilterFactory, NodeStandardFontDataFactory } = require$$0;
	DefaultCanvasFactory = NodeCanvasFactory;
	DefaultCMapReaderFactory = NodeCMapReaderFactory;
	DefaultFilterFactory = NodeFilterFactory;
	DefaultStandardFontDataFactory = NodeStandardFontDataFactory;
}

let createPDFNetworkStream;

{
	if (isNodeJS) {
		const { PDFNodeStream } = require$$1;

		createPDFNetworkStream = params => {
			return new PDFNodeStream(params);
		};
	} else {
		const { PDFNetworkStream } = require$$2;
		const { PDFFetchStream } = require$$3;

		createPDFNetworkStream = params => {
			return isValidFetchUrl(params.url)
			? new PDFFetchStream(params)
			: new PDFNetworkStream(params);
		};
	}
}

function getDocument(src) {
	if (typeof src === "string" || src instanceof URL) {
		src = { url: src };
	} else if (isArrayBuffer(src)) {
		src = { data: src };
	}

	if (typeof src !== "object") {
		throw new Error("Invalid parameter in getDocument, need parameter object.");
	}

	if (!src.url && !src.data && !src.range) {
		throw new Error("Invalid parameter object: need either .data, .range or .url");
	}

	const task = new PDFDocumentLoadingTask();
	const { docId } = task;
	const url = src.url ? getUrlProp(src.url) : null;
	const data = src.data ? getDataProp(src.data) : null;
	const httpHeaders = src.httpHeaders || null;
	const withCredentials = src.withCredentials === true;
	const password = src.password ?? null;

	const rangeTransport = src.range instanceof PDFDataRangeTransport
	? src.range
	: null;

	const rangeChunkSize = Number.isInteger(src.rangeChunkSize) && src.rangeChunkSize > 0
	? src.rangeChunkSize
	: DEFAULT_RANGE_CHUNK_SIZE;

	let worker = src.worker instanceof PDFWorker ? src.worker : null;
	const verbosity = src.verbosity;

	const docBaseUrl = typeof src.docBaseUrl === "string" && !isDataScheme(src.docBaseUrl)
	? src.docBaseUrl
	: null;

	const cMapUrl = typeof src.cMapUrl === "string" ? src.cMapUrl : null;
	const cMapPacked = src.cMapPacked !== false;
	const CMapReaderFactory = src.CMapReaderFactory || DefaultCMapReaderFactory;

	const standardFontDataUrl = typeof src.standardFontDataUrl === "string"
	? src.standardFontDataUrl
	: null;

	const StandardFontDataFactory = src.StandardFontDataFactory || DefaultStandardFontDataFactory;
	const ignoreErrors = src.stopAtErrors !== true;

	const maxImageSize = Number.isInteger(src.maxImageSize) && src.maxImageSize > -1
	? src.maxImageSize
	: -1;

	const isEvalSupported = src.isEvalSupported !== false;

	const isOffscreenCanvasSupported = typeof src.isOffscreenCanvasSupported === "boolean"
	? src.isOffscreenCanvasSupported
	: !isNodeJS;

	const canvasMaxAreaInBytes = Number.isInteger(src.canvasMaxAreaInBytes)
	? src.canvasMaxAreaInBytes
	: -1;

	const disableFontFace = typeof src.disableFontFace === "boolean"
	? src.disableFontFace
	: isNodeJS;

	const fontExtraProperties = src.fontExtraProperties === true;
	const enableXfa = src.enableXfa === true;
	const ownerDocument = src.ownerDocument || globalThis.document;
	const disableRange = src.disableRange === true;
	const disableStream = src.disableStream === true;
	const disableAutoFetch = src.disableAutoFetch === true;
	const pdfBug = src.pdfBug === true;

	const length = rangeTransport
	? rangeTransport.length
	: src.length ?? NaN;

	const useSystemFonts = typeof src.useSystemFonts === "boolean"
	? src.useSystemFonts
	: !isNodeJS && !disableFontFace;

	const useWorkerFetch = typeof src.useWorkerFetch === "boolean"
	? src.useWorkerFetch
	: CMapReaderFactory === DOMCMapReaderFactory && StandardFontDataFactory === DOMStandardFontDataFactory && isValidFetchUrl(cMapUrl, document.baseURI) && isValidFetchUrl(standardFontDataUrl, document.baseURI);

	const canvasFactory = src.canvasFactory || new DefaultCanvasFactory({ ownerDocument });
	const filterFactory = src.filterFactory || new DefaultFilterFactory({ docId, ownerDocument });
	const styleElement = null;
	setVerbosityLevel(verbosity);
	const transportFactory = { canvasFactory, filterFactory };

	if (!useWorkerFetch) {
		transportFactory.cMapReaderFactory = new CMapReaderFactory({
				baseUrl: cMapUrl,
				isCompressed: cMapPacked
			});

		transportFactory.standardFontDataFactory = new StandardFontDataFactory({ baseUrl: standardFontDataUrl });
	}

	if (!worker) {
		const workerParams = {
			verbosity,
			port: GlobalWorkerOptions.workerPort
		};

		worker = workerParams.port
		? PDFWorker.fromPort(workerParams)
		: new PDFWorker(workerParams);

		task._worker = worker;
	}

	const fetchDocParams = {
		docId,
		apiVersion: "ciri-custom",
		data,
		password,
		disableAutoFetch,
		rangeChunkSize,
		length,
		docBaseUrl,
		enableXfa,
		evaluatorOptions: {
			maxImageSize,
			disableFontFace,
			ignoreErrors,
			isEvalSupported,
			isOffscreenCanvasSupported,
			canvasMaxAreaInBytes,
			fontExtraProperties,
			useSystemFonts,
			cMapUrl: useWorkerFetch ? cMapUrl : null,
			standardFontDataUrl: useWorkerFetch ? standardFontDataUrl : null
		}
	};

	const transportParams = {
		ignoreErrors,
		isEvalSupported,
		disableFontFace,
		fontExtraProperties,
		enableXfa,
		ownerDocument,
		disableAutoFetch,
		pdfBug,
		styleElement
	};

	worker.promise.then(function () {
		if (task.destroyed) {
			throw new Error("Loading aborted");
		}

		const workerIdPromise = _fetchDocument(worker, fetchDocParams);

		const networkStreamPromise = new Promise(function (resolve) {
				let networkStream;

				if (rangeTransport) {
					networkStream = new PDFDataTransportStream({
							length,
							initialData: rangeTransport.initialData,
							progressiveDone: rangeTransport.progressiveDone,
							contentDispositionFilename: rangeTransport.contentDispositionFilename,
							disableRange,
							disableStream
						},
					rangeTransport);
				} else if (!data) {
					networkStream = createPDFNetworkStream({
						url,
						length,
						httpHeaders,
						withCredentials,
						rangeChunkSize,
						disableRange,
						disableStream
					});
				}

				resolve(networkStream);
			});

		return Promise.all([workerIdPromise, networkStreamPromise]).then(function ([workerId, networkStream]) {
			if (task.destroyed) {
				throw new Error("Loading aborted");
			}

			const messageHandler = new MessageHandler(docId, workerId, worker.port);
			const transport = new WorkerTransport(messageHandler, task, networkStream, transportParams, transportFactory);
			task._transport = transport;
			messageHandler.send("Ready", null);
		});
	}).catch(task._capability.reject);

	return task;
}

async function _fetchDocument(worker, source) {
	if (worker.destroyed) {
		throw new Error("Worker was destroyed");
	}

	const workerId = await worker.messageHandler.sendWithPromise("GetDocRequest", source, source.data ? [source.data.buffer] : null);

	if (worker.destroyed) {
		throw new Error("Worker was destroyed");
	}

	return workerId;
}

function getUrlProp(val) {
	if (val instanceof URL) {
		return val.href;
	}

	try {
		return new URL(val, window.location).href;
	} catch(ex) {
		if (isNodeJS && typeof val === "string") {
			return val;
		}
	}

	throw new Error("Invalid PDF url data: " + "either string or URL-object is expected in the url property.");
}

function getDataProp(val) {
	if (isNodeJS && typeof Buffer !== "undefined" && val instanceof Buffer) {
		deprecated("Please provide binary data as `Uint8Array`, rather than `Buffer`.");
		return new Uint8Array(val);
	}

	if (val instanceof Uint8Array && val.byteLength === val.buffer.byteLength) {
		return val;
	}

	if (typeof val === "string") {
		return stringToBytes(val);
	}

	if (typeof val === "object" && !isNaN(val?.length) || isArrayBuffer(val)) {
		return new Uint8Array(val);
	}

	throw new Error("Invalid PDF binary data: either TypedArray, " + "string, or array-like object is expected in the data property.");
}

class PDFDocumentLoadingTask {
	static #docId = 0;

	constructor() {
		this._capability = createPromiseCapability();
		this._transport = null;
		this._worker = null;
		this.docId = `d${PDFDocumentLoadingTask.#docId++}`;
		this.destroyed = false;
		this.onPassword = null;
		this.onProgress = null;
	}

	get promise() {
		return this._capability.promise;
	}

	async destroy() {
		this.destroyed = true;
		await this._transport?.destroy();
		this._transport = null;

		if (this._worker) {
			this._worker.destroy();
			this._worker = null;
		}
	}
}

class PDFDataRangeTransport {
	constructor(length, initialData, progressiveDone = false, contentDispositionFilename = null) {
		this.length = length;
		this.initialData = initialData;
		this.progressiveDone = progressiveDone;
		this.contentDispositionFilename = contentDispositionFilename;
		this._rangeListeners = [];
		this._progressListeners = [];
		this._progressiveReadListeners = [];
		this._progressiveDoneListeners = [];
		this._readyCapability = createPromiseCapability();
	}

	addRangeListener(listener) {
		this._rangeListeners.push(listener);
	}

	addProgressListener(listener) {
		this._progressListeners.push(listener);
	}

	addProgressiveReadListener(listener) {
		this._progressiveReadListeners.push(listener);
	}

	addProgressiveDoneListener(listener) {
		this._progressiveDoneListeners.push(listener);
	}

	onDataRange(begin, chunk) {
		for (const listener of this._rangeListeners) {
			listener(begin, chunk);
		}
	}

	onDataProgress(loaded, total) {
		this._readyCapability.promise.then(() => {
			for (const listener of this._progressListeners) {
				listener(loaded, total);
			}
		});
	}

	onDataProgressiveRead(chunk) {
		this._readyCapability.promise.then(() => {
			for (const listener of this._progressiveReadListeners) {
				listener(chunk);
			}
		});
	}

	onDataProgressiveDone() {
		this._readyCapability.promise.then(() => {
			for (const listener of this._progressiveDoneListeners) {
				listener();
			}
		});
	}

	transportReady() {
		this._readyCapability.resolve();
	}

	requestDataRange(begin, end) {
		unreachable("Abstract method PDFDataRangeTransport.requestDataRange");
	}

	abort() {
		
	}
}

class PDFDocumentProxy {
	constructor(pdfInfo, transport) {
		this._pdfInfo = pdfInfo;
		this._transport = transport;
	}

	get annotationStorage() {
		return this._transport.annotationStorage;
	}

	get filterFactory() {
		return this._transport.filterFactory;
	}

	get numPages() {
		return this._pdfInfo.numPages;
	}

	get fingerprints() {
		return this._pdfInfo.fingerprints;
	}

	get isPureXfa() {
		return shadow(this, "isPureXfa", !!this._transport._htmlForXfa);
	}

	get allXfaHtml() {
		return this._transport._htmlForXfa;
	}

	getPage(pageNumber) {
		return this._transport.getPage(pageNumber);
	}

	getPageIndex(ref) {
		return this._transport.getPageIndex(ref);
	}

	getDestinations() {
		return this._transport.getDestinations();
	}

	getDestination(id) {
		return this._transport.getDestination(id);
	}

	getPageLabels() {
		return this._transport.getPageLabels();
	}

	getPageLayout() {
		return this._transport.getPageLayout();
	}

	getPageMode() {
		return this._transport.getPageMode();
	}

	getViewerPreferences() {
		return this._transport.getViewerPreferences();
	}

	getOpenAction() {
		return this._transport.getOpenAction();
	}

	getAttachments() {
		return this._transport.getAttachments();
	}

	getJavaScript() {
		return this._transport.getJavaScript();
	}

	getJSActions() {
		return this._transport.getDocJSActions();
	}

	getOutline() {
		return this._transport.getOutline();
	}

	getOptionalContentConfig() {
		return this._transport.getOptionalContentConfig();
	}

	getPermissions() {
		return this._transport.getPermissions();
	}

	getMetadata() {
		return this._transport.getMetadata();
	}

	getMarkInfo() {
		return this._transport.getMarkInfo();
	}

	getData() {
		return this._transport.getData();
	}

	saveDocument() {
		return this._transport.saveDocument();
	}

	getDownloadInfo() {
		return this._transport.downloadInfoCapability.promise;
	}

	cleanup(keepLoadedFonts = false) {
		return this._transport.startCleanup(keepLoadedFonts || this.isPureXfa);
	}

	destroy() {
		return this.loadingTask.destroy();
	}

	get loadingParams() {
		return this._transport.loadingParams;
	}

	get loadingTask() {
		return this._transport.loadingTask;
	}

	getFieldObjects() {
		return this._transport.getFieldObjects();
	}

	hasJSActions() {
		return this._transport.hasJSActions();
	}

	getCalculationOrderIds() {
		return this._transport.getCalculationOrderIds();
	}
}

class PDFPageProxy {
	#delayedCleanupTimeout = null;
	#pendingCleanup = false;

	constructor(pageIndex, pageInfo, transport, pdfBug = false) {
		this._pageIndex = pageIndex;
		this._pageInfo = pageInfo;
		this._transport = transport;
		this._stats = pdfBug ? new StatTimer() : null;
		this._pdfBug = pdfBug;
		this.commonObjs = transport.commonObjs;
		this.objs = new PDFObjects();
		this._maybeCleanupAfterRender = false;
		this._intentStates = new Map();
		this.destroyed = false;
	}

	get pageNumber() {
		return this._pageIndex + 1;
	}

	get rotate() {
		return this._pageInfo.rotate;
	}

	get ref() {
		return this._pageInfo.ref;
	}

	get userUnit() {
		return this._pageInfo.userUnit;
	}

	get view() {
		return this._pageInfo.view;
	}

	getViewport({ scale, rotation = this.rotate, offsetX = 0, offsetY = 0, dontFlip = false } = {}) {
		return new PageViewport({
				viewBox: this.view,
				scale,
				rotation,
				offsetX,
				offsetY,
				dontFlip
			});
	}

	getAnnotations({ intent = "display" } = {}) {
		const intentArgs = this._transport.getRenderingIntent(intent);
		return this._transport.getAnnotations(this._pageIndex, intentArgs.renderingIntent);
	}

	getJSActions() {
		return this._transport.getPageJSActions(this._pageIndex);
	}

	get isPureXfa() {
		return shadow(this, "isPureXfa", !!this._transport._htmlForXfa);
	}

	async getXfa() {
		return this._transport._htmlForXfa?.children[this._pageIndex] || null;
	}

	render({ canvasContext, viewport, intent = "display", annotationMode = AnnotationMode.ENABLE, transform = null, background = null, optionalContentConfigPromise = null, annotationCanvasMap = null, pageColors = null, printAnnotationStorage = null }) {
		if (arguments[0]?.canvasFactory) {
			throw new Error("render no longer accepts the `canvasFactory`-option, " + "please pass it to the `getDocument`-function instead.");
		}

		this._stats?.time("Overall");
		const intentArgs = this._transport.getRenderingIntent(intent, annotationMode, printAnnotationStorage);
		this.#pendingCleanup = false;
		this.#abortDelayedCleanup();

		if (!optionalContentConfigPromise) {
			optionalContentConfigPromise = this._transport.getOptionalContentConfig();
		}

		let intentState = this._intentStates.get(intentArgs.cacheKey);

		if (!intentState) {
			intentState = Object.create(null);
			this._intentStates.set(intentArgs.cacheKey, intentState);
		}

		if (intentState.streamReaderCancelTimeout) {
			clearTimeout(intentState.streamReaderCancelTimeout);
			intentState.streamReaderCancelTimeout = null;
		}

		const intentPrint = !!(intentArgs.renderingIntent & RenderingIntentFlag.PRINT);

		if (!intentState.displayReadyCapability) {
			intentState.displayReadyCapability = createPromiseCapability();

			intentState.operatorList = {
				fnArray: [],
				argsArray: [],
				lastChunk: false,
				separateAnnots: null
			};

			this._stats?.time("Page Request");
			this._pumpOperatorList(intentArgs);
		}

		const complete = error => {
			intentState.renderTasks.delete(internalRenderTask);

			if (this._maybeCleanupAfterRender || intentPrint) {
				this.#pendingCleanup = true;
			}

			this.#tryCleanup(!intentPrint);

			if (error) {
				internalRenderTask.capability.reject(error);

				this._abortOperatorList({
					intentState,
					reason: error instanceof Error ? error : new Error(error)
				});
			} else {
				internalRenderTask.capability.resolve();
			}

			this._stats?.timeEnd("Rendering");
			this._stats?.timeEnd("Overall");
		};

		const internalRenderTask = new InternalRenderTask({
				callback: complete,
				params: {
					canvasContext,
					viewport,
					transform,
					background
				},
				objs: this.objs,
				commonObjs: this.commonObjs,
				annotationCanvasMap,
				operatorList: intentState.operatorList,
				pageIndex: this._pageIndex,
				canvasFactory: this._transport.canvasFactory,
				filterFactory: this._transport.filterFactory,
				useRequestAnimationFrame: !intentPrint,
				pdfBug: this._pdfBug,
				pageColors
			});

		(intentState.renderTasks ||= new Set()).add(internalRenderTask);
		const renderTask = internalRenderTask.task;

		Promise.all([intentState.displayReadyCapability.promise, optionalContentConfigPromise]).then(([transparency, optionalContentConfig]) => {
			if (this.#pendingCleanup) {
				complete();
				return;
			}

			this._stats?.time("Rendering");
			internalRenderTask.initializeGraphics({ transparency, optionalContentConfig });
			internalRenderTask.operatorListChanged();
		}).catch(complete);

		return renderTask;
	}

	getOperatorList({ intent = "display", annotationMode = AnnotationMode.ENABLE, printAnnotationStorage = null } = {}) {
		function operatorListChanged() {
			if (intentState.operatorList.lastChunk) {
				intentState.opListReadCapability.resolve(intentState.operatorList);
				intentState.renderTasks.delete(opListTask);
			}
		}

		const intentArgs = this._transport.getRenderingIntent(intent, annotationMode, printAnnotationStorage, true);
		let intentState = this._intentStates.get(intentArgs.cacheKey);

		if (!intentState) {
			intentState = Object.create(null);
			this._intentStates.set(intentArgs.cacheKey, intentState);
		}

		let opListTask;

		if (!intentState.opListReadCapability) {
			opListTask = Object.create(null);
			opListTask.operatorListChanged = operatorListChanged;
			intentState.opListReadCapability = createPromiseCapability();
			(intentState.renderTasks ||= new Set()).add(opListTask);

			intentState.operatorList = {
				fnArray: [],
				argsArray: [],
				lastChunk: false,
				separateAnnots: null
			};

			this._stats?.time("Page Request");
			this._pumpOperatorList(intentArgs);
		}

		return intentState.opListReadCapability.promise;
	}

	streamTextContent({ includeMarkedContent = false, disableNormalization = false } = {}) {
		const TEXT_CONTENT_CHUNK_SIZE = 100;

		return this._transport.messageHandler.sendWithStream(
			"GetTextContent",
			{
				pageIndex: this._pageIndex,
				includeMarkedContent: includeMarkedContent === true,
				disableNormalization: disableNormalization === true
			},
			{
				highWaterMark: TEXT_CONTENT_CHUNK_SIZE,
				size(textContent) {
					return textContent.items.length;
				}
			}
		);
	}

	getTextContent(params = {}) {
		if (this._transport._htmlForXfa) {
			return this.getXfa().then(xfa => {
				return XfaText.textContent(xfa);
			});
		}

		const readableStream = this.streamTextContent(params);

		return new Promise(function (resolve, reject) {
				function pump() {
					reader.read().then(
						function ({ value, done }) {
							if (done) {
								resolve(textContent);
								return;
							}

							Object.assign(textContent.styles, value.styles);
							textContent.items.push(...value.items);
							pump();
						},
						reject
					);
				}

				const reader = readableStream.getReader();
				const textContent = { items: [], styles: Object.create(null) };
				pump();
			});
	}

	getStructTree() {
		return this._transport.getStructTree(this._pageIndex);
	}

	_destroy() {
		this.destroyed = true;
		const waitOn = [];

		for (const intentState of this._intentStates.values()) {
			this._abortOperatorList({
				intentState,
				reason: new Error("Page was destroyed."),
				force: true
			});

			if (intentState.opListReadCapability) {
				continue;
			}

			for (const internalRenderTask of intentState.renderTasks) {
				waitOn.push(internalRenderTask.completed);
				internalRenderTask.cancel();
			}
		}

		this.objs.clear();
		this.#pendingCleanup = false;
		this.#abortDelayedCleanup();
		return Promise.all(waitOn);
	}

	cleanup(resetStats = false) {
		this.#pendingCleanup = true;
		const success = this.#tryCleanup(false);

		if (resetStats && success) {
			this._stats &&= new StatTimer();
		}

		return success;
	}

	#tryCleanup(delayed = false) {
		this.#abortDelayedCleanup();

		if (!this.#pendingCleanup) {
			return false;
		}

		if (delayed) {
			this.#delayedCleanupTimeout = setTimeout(
				() => {
					this.#delayedCleanupTimeout = null;
					this.#tryCleanup(false);
				},
				DELAYED_CLEANUP_TIMEOUT
			);

			return false;
		}

		for (const { renderTasks, operatorList } of this._intentStates.values()) {
			if (renderTasks.size > 0 || !operatorList.lastChunk) {
				return false;
			}
		}

		this._intentStates.clear();
		this.objs.clear();
		this.#pendingCleanup = false;
		return true;
	}

	#abortDelayedCleanup() {
		if (this.#delayedCleanupTimeout) {
			clearTimeout(this.#delayedCleanupTimeout);
			this.#delayedCleanupTimeout = null;
		}
	}

	_startRenderPage(transparency, cacheKey) {
		const intentState = this._intentStates.get(cacheKey);

		if (!intentState) {
			return;
		}

		this._stats?.timeEnd("Page Request");
		intentState.displayReadyCapability?.resolve(transparency);
	}

	_renderPageChunk(operatorListChunk, intentState) {
		for (let i = 0, ii = operatorListChunk.length; i < ii; i++) {
			intentState.operatorList.fnArray.push(operatorListChunk.fnArray[i]);
			intentState.operatorList.argsArray.push(operatorListChunk.argsArray[i]);
		}

		intentState.operatorList.lastChunk = operatorListChunk.lastChunk;
		intentState.operatorList.separateAnnots = operatorListChunk.separateAnnots;

		for (const internalRenderTask of intentState.renderTasks) {
			internalRenderTask.operatorListChanged();
		}

		if (operatorListChunk.lastChunk) {
			this.#tryCleanup(true);
		}
	}

	_pumpOperatorList({ renderingIntent, cacheKey, annotationStorageMap }) {
		const readableStream = this._transport.messageHandler.sendWithStream("GetOperatorList", {
			pageIndex: this._pageIndex,
			intent: renderingIntent,
			cacheKey,
			annotationStorage: annotationStorageMap
		});

		const reader = readableStream.getReader();
		const intentState = this._intentStates.get(cacheKey);
		intentState.streamReader = reader;

		const pump = () => {
			reader.read().then(
				({ value, done }) => {
					if (done) {
						intentState.streamReader = null;
						return;
					}

					if (this._transport.destroyed) {
						return;
					}

					this._renderPageChunk(value, intentState);
					pump();
				},
				reason => {
					intentState.streamReader = null;

					if (this._transport.destroyed) {
						return;
					}

					if (intentState.operatorList) {
						intentState.operatorList.lastChunk = true;

						for (const internalRenderTask of intentState.renderTasks) {
							internalRenderTask.operatorListChanged();
						}

						this.#tryCleanup(true);
					}

					if (intentState.displayReadyCapability) {
						intentState.displayReadyCapability.reject(reason);
					} else if (intentState.opListReadCapability) {
						intentState.opListReadCapability.reject(reason);
					} else {
						throw reason;
					}
				}
			);
		};

		pump();
	}

	_abortOperatorList({ intentState, reason, force = false }) {
		if (!intentState.streamReader) {
			return;
		}

		if (intentState.streamReaderCancelTimeout) {
			clearTimeout(intentState.streamReaderCancelTimeout);
			intentState.streamReaderCancelTimeout = null;
		}

		if (!force) {
			if (intentState.renderTasks.size > 0) {
				return;
			}

			if (reason instanceof RenderingCancelledException) {
				let delay = RENDERING_CANCELLED_TIMEOUT;

				if (reason.extraDelay > 0 && reason.extraDelay < 1000) {
					delay += reason.extraDelay;
				}

				intentState.streamReaderCancelTimeout = setTimeout(
					() => {
						intentState.streamReaderCancelTimeout = null;
						this._abortOperatorList({ intentState, reason, force: true });
					},
					delay
				);

				return;
			}
		}

		intentState.streamReader.cancel(new AbortException(reason.message)).catch(() => {
			
		});

		intentState.streamReader = null;

		if (this._transport.destroyed) {
			return;
		}

		for (const [curCacheKey, curIntentState] of this._intentStates) {
			if (curIntentState === intentState) {
				this._intentStates.delete(curCacheKey);
				break;
			}
		}

		this.cleanup();
	}

	get stats() {
		return this._stats;
	}
}

class LoopbackPort {
	#listeners = new Set();
	#deferred = Promise.resolve();

	postMessage(obj, transfer) {
		const event = {
			data: structuredClone(obj, transfer ? { transfer } : null)
		};

		this.#deferred.then(() => {
			for (const listener of this.#listeners) {
				listener.call(this, event);
			}
		});
	}

	addEventListener(name, listener) {
		this.#listeners.add(listener);
	}

	removeEventListener(name, listener) {
		this.#listeners.delete(listener);
	}

	terminate() {
		this.#listeners.clear();
	}
}

const PDFWorkerUtil = {
	isWorkerDisabled: false,
	fallbackWorkerSrc: null,
	fakeWorkerId: 0
};

{
	if (isNodeJS && typeof require === "function") {
		PDFWorkerUtil.isWorkerDisabled = true;
		PDFWorkerUtil.fallbackWorkerSrc = "./pdf.worker.js";
	} else if (typeof document === "object") {
		const pdfjsFilePath = document?.currentScript?.src;

		if (pdfjsFilePath) {
			PDFWorkerUtil.fallbackWorkerSrc = pdfjsFilePath.replace(/(\.(?:min\.)?js)(\?.*)?$/i, ".worker$1$2");
		}
	}

	PDFWorkerUtil.isSameOrigin = function (baseUrl, otherUrl) {
		let base;

		try {
			base = new URL(baseUrl);

			if (!base.origin || base.origin === "null") {
				return false;
			}
		} catch(e) {
			return false;
		}

		const other = new URL(otherUrl, base);
		return base.origin === other.origin;
	};

	PDFWorkerUtil.createCDNWrapper = function (url) {
		const wrapper = `importScripts("${url}");`;
		return URL.createObjectURL(new Blob([wrapper]));
	};
}

class PDFWorker {
	static #workerPorts = new WeakMap();

	constructor({ name = null, port = null, verbosity = getVerbosityLevel() } = {}) {
		if (port && PDFWorker.#workerPorts.has(port)) {
			throw new Error("Cannot use more than one PDFWorker per port.");
		}

		this.name = name;
		this.destroyed = false;
		this.verbosity = verbosity;
		this._readyCapability = createPromiseCapability();
		this._port = null;
		this._webWorker = null;
		this._messageHandler = null;

		if (port) {
			PDFWorker.#workerPorts.set(port, this);
			this._initializeFromPort(port);
			return;
		}

		this._initialize();
	}

	get promise() {
		return this._readyCapability.promise;
	}

	get port() {
		return this._port;
	}

	get messageHandler() {
		return this._messageHandler;
	}

	_initializeFromPort(port) {
		this._port = port;
		this._messageHandler = new MessageHandler("main", "worker", port);

		this._messageHandler.on("ready", function () {
			
		});

		this._readyCapability.resolve();
		this._messageHandler.send("configure", { verbosity: this.verbosity });
	}

	_initialize() {
		if (!PDFWorkerUtil.isWorkerDisabled && !PDFWorker._mainThreadWorkerMessageHandler) {
			let { workerSrc } = PDFWorker;

			try {
				if (!PDFWorkerUtil.isSameOrigin(window.location.href, workerSrc)) {
					workerSrc = PDFWorkerUtil.createCDNWrapper(new URL(workerSrc, window.location).href);
				}

				const worker = new Worker(workerSrc);
				const messageHandler = new MessageHandler("main", "worker", worker);

				const terminateEarly = () => {
					worker.removeEventListener("error", onWorkerError);
					messageHandler.destroy();
					worker.terminate();

					if (this.destroyed) {
						this._readyCapability.reject(new Error("Worker was destroyed"));
					} else {
						this._setupFakeWorker();
					}
				};

				const onWorkerError = () => {
					if (!this._webWorker) {
						terminateEarly();
					}
				};

				worker.addEventListener("error", onWorkerError);

				messageHandler.on("test", data => {
					worker.removeEventListener("error", onWorkerError);

					if (this.destroyed) {
						terminateEarly();
						return;
					}

					if (data) {
						this._messageHandler = messageHandler;
						this._port = worker;
						this._webWorker = worker;
						this._readyCapability.resolve();
						messageHandler.send("configure", { verbosity: this.verbosity });
					} else {
						this._setupFakeWorker();
						messageHandler.destroy();
						worker.terminate();
					}
				});

				messageHandler.on("ready", data => {
					worker.removeEventListener("error", onWorkerError);

					if (this.destroyed) {
						terminateEarly();
						return;
					}

					try {
						sendTest();
					} catch(e) {
						this._setupFakeWorker();
					}
				});

				const sendTest = () => {
					const testObj = new Uint8Array();
					messageHandler.send("test", testObj, [testObj.buffer]);
				};

				sendTest();
				return;
			} catch(e) {
				info("The worker has been disabled.");
			}
		}

		this._setupFakeWorker();
	}

	_setupFakeWorker() {
		if (!PDFWorkerUtil.isWorkerDisabled) {
			warn("Setting up fake worker.");
			PDFWorkerUtil.isWorkerDisabled = true;
		}

		PDFWorker._setupFakeWorkerGlobal.then(WorkerMessageHandler => {
			if (this.destroyed) {
				this._readyCapability.reject(new Error("Worker was destroyed"));
				return;
			}

			const port = new LoopbackPort();
			this._port = port;
			const id = `fake${PDFWorkerUtil.fakeWorkerId++}`;
			const workerHandler = new MessageHandler(id + "_worker", id, port);
			WorkerMessageHandler.setup(workerHandler, port);
			const messageHandler = new MessageHandler(id, id + "_worker", port);
			this._messageHandler = messageHandler;
			this._readyCapability.resolve();
			messageHandler.send("configure", { verbosity: this.verbosity });
		}).catch(reason => {
			this._readyCapability.reject(new Error(`Setting up fake worker failed: "${reason.message}".`));
		});
	}

	destroy() {
		this.destroyed = true;

		if (this._webWorker) {
			this._webWorker.terminate();
			this._webWorker = null;
		}

		PDFWorker.#workerPorts.delete(this._port);
		this._port = null;

		if (this._messageHandler) {
			this._messageHandler.destroy();
			this._messageHandler = null;
		}
	}

	static fromPort(params) {
		if (!params?.port) {
			throw new Error("PDFWorker.fromPort - invalid method signature.");
		}

		if (this.#workerPorts.has(params.port)) {
			return this.#workerPorts.get(params.port);
		}

		return new PDFWorker(params);
	}

	static get workerSrc() {
		if (GlobalWorkerOptions.workerSrc) {
			return GlobalWorkerOptions.workerSrc;
		}

		if (PDFWorkerUtil.fallbackWorkerSrc !== null) {
			if (!isNodeJS) {
				deprecated('No "GlobalWorkerOptions.workerSrc" specified.');
			}

			return PDFWorkerUtil.fallbackWorkerSrc;
		}

		throw new Error('No "GlobalWorkerOptions.workerSrc" specified.');
	}

	static get _mainThreadWorkerMessageHandler() {
		try {
			return globalThis.pdfjsWorker?.WorkerMessageHandler || null;
		} catch(ex) {
			return null;
		}
	}

	static get _setupFakeWorkerGlobal() {
		const loader = async () => {
			const mainWorkerMessageHandler = this._mainThreadWorkerMessageHandler;

			if (mainWorkerMessageHandler) {
				return mainWorkerMessageHandler;
			}

			if (isNodeJS && typeof require === "function") {
				const worker = require(this.workerSrc);
				return worker.WorkerMessageHandler;
			}

			await loadScript(this.workerSrc);
			return window.pdfjsWorker.WorkerMessageHandler;
		};

		return shadow(this, "_setupFakeWorkerGlobal", loader());
	}
}

class WorkerTransport {
	#methodPromises = new Map();
	#pageCache = new Map();
	#pagePromises = new Map();

	constructor(messageHandler, loadingTask, networkStream, params, factory) {
		this.messageHandler = messageHandler;
		this.loadingTask = loadingTask;
		this.commonObjs = new PDFObjects();

		this.fontLoader = new FontLoader({
				ownerDocument: params.ownerDocument,
				styleElement: params.styleElement
			});

		this._params = params;
		this.canvasFactory = factory.canvasFactory;
		this.filterFactory = factory.filterFactory;
		this.cMapReaderFactory = factory.cMapReaderFactory;
		this.standardFontDataFactory = factory.standardFontDataFactory;
		this.destroyed = false;
		this.destroyCapability = null;
		this._passwordCapability = null;
		this._networkStream = networkStream;
		this._fullReader = null;
		this._lastProgress = null;
		this.downloadInfoCapability = createPromiseCapability();
		this.setupMessageHandler();
	}

	#cacheSimpleMethod(name, data = null) {
		const cachedPromise = this.#methodPromises.get(name);

		if (cachedPromise) {
			return cachedPromise;
		}

		const promise = this.messageHandler.sendWithPromise(name, data);
		this.#methodPromises.set(name, promise);
		return promise;
	}

	get annotationStorage() {
		return shadow(this, "annotationStorage", new AnnotationStorage());
	}

	getRenderingIntent(intent, annotationMode = AnnotationMode.ENABLE, printAnnotationStorage = null, isOpList = false) {
		let renderingIntent = RenderingIntentFlag.DISPLAY;
		let annotationMap = null;

		switch (intent) {
			case "any":
				renderingIntent = RenderingIntentFlag.ANY;
				break;
			case "display":
				break;
			case "print":
				renderingIntent = RenderingIntentFlag.PRINT;
				break;
			default:
				warn(`getRenderingIntent - invalid intent: ${intent}`);
		}

		switch (annotationMode) {
			case AnnotationMode.DISABLE:
				renderingIntent += RenderingIntentFlag.ANNOTATIONS_DISABLE;
				break;
			case AnnotationMode.ENABLE:
				break;
			case AnnotationMode.ENABLE_FORMS:
				renderingIntent += RenderingIntentFlag.ANNOTATIONS_FORMS;
				break;
			case AnnotationMode.ENABLE_STORAGE:
				renderingIntent += RenderingIntentFlag.ANNOTATIONS_STORAGE;
				const annotationStorage = renderingIntent & RenderingIntentFlag.PRINT && printAnnotationStorage instanceof PrintAnnotationStorage
				? printAnnotationStorage
				: this.annotationStorage;
				annotationMap = annotationStorage.serializable;
				break;
			default:
				warn(`getRenderingIntent - invalid annotationMode: ${annotationMode}`);
		}

		if (isOpList) {
			renderingIntent += RenderingIntentFlag.OPLIST;
		}

		return {
			renderingIntent,
			cacheKey: `${renderingIntent}_${AnnotationStorage.getHash(annotationMap)}`,
			annotationStorageMap: annotationMap
		};
	}

	destroy() {
		if (this.destroyCapability) {
			return this.destroyCapability.promise;
		}

		this.destroyed = true;
		this.destroyCapability = createPromiseCapability();

		if (this._passwordCapability) {
			this._passwordCapability.reject(new Error("Worker was destroyed during onPassword callback"));
		}

		const waitOn = [];

		for (const page of this.#pageCache.values()) {
			waitOn.push(page._destroy());
		}

		this.#pageCache.clear();
		this.#pagePromises.clear();

		if (this.hasOwnProperty("annotationStorage")) {
			this.annotationStorage.resetModified();
		}

		const terminated = this.messageHandler.sendWithPromise("Terminate", null);
		waitOn.push(terminated);

		Promise.all(waitOn).then(
			() => {
				this.commonObjs.clear();
				this.fontLoader.clear();
				this.#methodPromises.clear();
				this.filterFactory.destroy();

				if (this._networkStream) {
					this._networkStream.cancelAllRequests(new AbortException("Worker was terminated."));
				}

				if (this.messageHandler) {
					this.messageHandler.destroy();
					this.messageHandler = null;
				}

				this.destroyCapability.resolve();
			},
			this.destroyCapability.reject
		);

		return this.destroyCapability.promise;
	}

	setupMessageHandler() {
		const { messageHandler, loadingTask } = this;

		messageHandler.on("GetReader", (data, sink) => {
			assert(this._networkStream, "GetReader - no `IPDFStream` instance available.");
			this._fullReader = this._networkStream.getFullReader();

			this._fullReader.onProgress = evt => {
				this._lastProgress = { loaded: evt.loaded, total: evt.total };
			};

			sink.onPull = () => {
				this._fullReader.read().then(function ({ value, done }) {
					if (done) {
						sink.close();
						return;
					}

					assert(value instanceof ArrayBuffer, "GetReader - expected an ArrayBuffer.");
					sink.enqueue(new Uint8Array(value), 1, [value]);
				}).catch(reason => {
					sink.error(reason);
				});
			};

			sink.onCancel = reason => {
				this._fullReader.cancel(reason);

				sink.ready.catch(readyReason => {
					if (this.destroyed) {
						return;
					}

					throw readyReason;
				});
			};
		});

		messageHandler.on("ReaderHeadersReady", data => {
			const headersCapability = createPromiseCapability();
			const fullReader = this._fullReader;

			fullReader.headersReady.then(
				() => {
					if (!fullReader.isStreamingSupported || !fullReader.isRangeSupported) {
						if (this._lastProgress) {
							loadingTask.onProgress?.(this._lastProgress);
						}

						fullReader.onProgress = evt => {
							loadingTask.onProgress?.({ loaded: evt.loaded, total: evt.total });
						};
					}

					headersCapability.resolve({
						isStreamingSupported: fullReader.isStreamingSupported,
						isRangeSupported: fullReader.isRangeSupported,
						contentLength: fullReader.contentLength
					});
				},
				headersCapability.reject
			);

			return headersCapability.promise;
		});

		messageHandler.on("GetRangeReader", (data, sink) => {
			assert(this._networkStream, "GetRangeReader - no `IPDFStream` instance available.");
			const rangeReader = this._networkStream.getRangeReader(data.begin, data.end);

			if (!rangeReader) {
				sink.close();
				return;
			}

			sink.onPull = () => {
				rangeReader.read().then(function ({ value, done }) {
					if (done) {
						sink.close();
						return;
					}

					assert(value instanceof ArrayBuffer, "GetRangeReader - expected an ArrayBuffer.");
					sink.enqueue(new Uint8Array(value), 1, [value]);
				}).catch(reason => {
					sink.error(reason);
				});
			};

			sink.onCancel = reason => {
				rangeReader.cancel(reason);

				sink.ready.catch(readyReason => {
					if (this.destroyed) {
						return;
					}

					throw readyReason;
				});
			};
		});

		messageHandler.on("GetDoc", ({ pdfInfo }) => {
			this._numPages = pdfInfo.numPages;
			this._htmlForXfa = pdfInfo.htmlForXfa;
			delete pdfInfo.htmlForXfa;
			loadingTask._capability.resolve(new PDFDocumentProxy(pdfInfo, this));
		});

		messageHandler.on("DocException", function (ex) {
			let reason;

			switch (ex.name) {
				case "PasswordException":
					reason = new PasswordException(ex.message, ex.code);
					break;
				case "InvalidPDFException":
					reason = new InvalidPDFException(ex.message);
					break;
				case "MissingPDFException":
					reason = new MissingPDFException(ex.message);
					break;
				case "UnexpectedResponseException":
					reason = new UnexpectedResponseException(ex.message, ex.status);
					break;
				case "UnknownErrorException":
					reason = new UnknownErrorException(ex.message, ex.details);
					break;
				default:
					unreachable("DocException - expected a valid Error.");
			}

			loadingTask._capability.reject(reason);
		});

		messageHandler.on("PasswordRequest", exception => {
			this._passwordCapability = createPromiseCapability();

			if (loadingTask.onPassword) {
				const updatePassword = password => {
					if (password instanceof Error) {
						this._passwordCapability.reject(password);
					} else {
						this._passwordCapability.resolve({ password });
					}
				};

				try {
					loadingTask.onPassword(updatePassword, exception.code);
				} catch(ex) {
					this._passwordCapability.reject(ex);
				}
			} else {
				this._passwordCapability.reject(new PasswordException(exception.message, exception.code));
			}

			return this._passwordCapability.promise;
		});

		messageHandler.on("DataLoaded", data => {
			loadingTask.onProgress?.({ loaded: data.length, total: data.length });
			this.downloadInfoCapability.resolve(data);
		});

		messageHandler.on("StartRenderPage", data => {
			if (this.destroyed) {
				return;
			}

			const page = this.#pageCache.get(data.pageIndex);
			page._startRenderPage(data.transparency, data.cacheKey);
		});

		messageHandler.on("commonobj", ([id, type, exportedData]) => {
			if (this.destroyed) {
				return;
			}

			if (this.commonObjs.has(id)) {
				return;
			}

			switch (type) {
				case "Font":
					const params = this._params;
					if ("error" in exportedData) {
						const exportedError = exportedData.error;
						warn(`Error during font loading: ${exportedError}`);
						this.commonObjs.resolve(id, exportedError);
						break;
					}
					const inspectFont = params.pdfBug && globalThis.FontInspector?.enabled
					? (font, url) => globalThis.FontInspector.fontAdded(font, url)
					: null;
					const font = new FontFaceObject(exportedData,
					{
							isEvalSupported: params.isEvalSupported,
							disableFontFace: params.disableFontFace,
							ignoreErrors: params.ignoreErrors,
							inspectFont
						});
					this.fontLoader.bind(font).catch(reason => {
						return messageHandler.sendWithPromise("FontFallback", { id });
					}).finally(() => {
						if (!params.fontExtraProperties && font.data) {
							font.data = null;
						}

						this.commonObjs.resolve(id, font);
					});
					break;
				case "FontPath":
				case "Image":
				case "Pattern":
					this.commonObjs.resolve(id, exportedData);
					break;
				default:
					throw new Error(`Got unknown common object type ${type}`);
			}
		});

		messageHandler.on("obj", ([id, pageIndex, type, imageData]) => {
			if (this.destroyed) {
				return;
			}

			const pageProxy = this.#pageCache.get(pageIndex);

			if (pageProxy.objs.has(id)) {
				return;
			}

			switch (type) {
				case "Image":
					pageProxy.objs.resolve(id, imageData);
					if (imageData) {
						let length;

						if (imageData.bitmap) {
							const { width, height } = imageData;
							length = width * height * 4;
						} else {
							length = imageData.data?.length || 0;
						}

						if (length > MAX_IMAGE_SIZE_TO_CACHE) {
							pageProxy._maybeCleanupAfterRender = true;
						}
					}
					break;
				case "Pattern":
					pageProxy.objs.resolve(id, imageData);
					break;
				default:
					throw new Error(`Got unknown object type ${type}`);
			}
		});

		messageHandler.on("DocProgress", data => {
			if (this.destroyed) {
				return;
			}

			loadingTask.onProgress?.({ loaded: data.loaded, total: data.total });
		});

		messageHandler.on("FetchBuiltInCMap", data => {
			if (this.destroyed) {
				return Promise.reject(new Error("Worker was destroyed."));
			}

			if (!this.cMapReaderFactory) {
				return Promise.reject(new Error("CMapReaderFactory not initialized, see the `useWorkerFetch` parameter."));
			}

			return this.cMapReaderFactory.fetch(data);
		});

		messageHandler.on("FetchStandardFontData", data => {
			if (this.destroyed) {
				return Promise.reject(new Error("Worker was destroyed."));
			}

			if (!this.standardFontDataFactory) {
				return Promise.reject(new Error("StandardFontDataFactory not initialized, see the `useWorkerFetch` parameter."));
			}

			return this.standardFontDataFactory.fetch(data);
		});
	}

	getData() {
		return this.messageHandler.sendWithPromise("GetData", null);
	}

	saveDocument() {
		if (this.annotationStorage.size <= 0) {
			warn("saveDocument called while `annotationStorage` is empty, " + "please use the getData-method instead.");
		}

		return this.messageHandler.sendWithPromise("SaveDocument", {
			isPureXfa: !!this._htmlForXfa,
			numPages: this._numPages,
			annotationStorage: this.annotationStorage.serializable,
			filename: this._fullReader?.filename ?? null
		}).finally(() => {
			this.annotationStorage.resetModified();
		});
	}

	getPage(pageNumber) {
		if (!Number.isInteger(pageNumber) || pageNumber <= 0 || pageNumber > this._numPages) {
			return Promise.reject(new Error("Invalid page request."));
		}

		const pageIndex = pageNumber - 1,
			cachedPromise = this.#pagePromises.get(pageIndex);

		if (cachedPromise) {
			return cachedPromise;
		}

		const promise = this.messageHandler.sendWithPromise("GetPage", { pageIndex }).then(pageInfo => {
			if (this.destroyed) {
				throw new Error("Transport destroyed");
			}

			const page = new PDFPageProxy(pageIndex, pageInfo, this, this._params.pdfBug);
			this.#pageCache.set(pageIndex, page);
			return page;
		});

		this.#pagePromises.set(pageIndex, promise);
		return promise;
	}

	getPageIndex(ref) {
		if (typeof ref !== "object" || ref === null || !Number.isInteger(ref.num) || ref.num < 0 || !Number.isInteger(ref.gen) || ref.gen < 0) {
			return Promise.reject(new Error("Invalid pageIndex request."));
		}

		return this.messageHandler.sendWithPromise("GetPageIndex", { num: ref.num, gen: ref.gen });
	}

	getAnnotations(pageIndex, intent) {
		return this.messageHandler.sendWithPromise("GetAnnotations", { pageIndex, intent });
	}

	getFieldObjects() {
		return this.#cacheSimpleMethod("GetFieldObjects");
	}

	hasJSActions() {
		return this.#cacheSimpleMethod("HasJSActions");
	}

	getCalculationOrderIds() {
		return this.messageHandler.sendWithPromise("GetCalculationOrderIds", null);
	}

	getDestinations() {
		return this.messageHandler.sendWithPromise("GetDestinations", null);
	}

	getDestination(id) {
		if (typeof id !== "string") {
			return Promise.reject(new Error("Invalid destination request."));
		}

		return this.messageHandler.sendWithPromise("GetDestination", { id });
	}

	getPageLabels() {
		return this.messageHandler.sendWithPromise("GetPageLabels", null);
	}

	getPageLayout() {
		return this.messageHandler.sendWithPromise("GetPageLayout", null);
	}

	getPageMode() {
		return this.messageHandler.sendWithPromise("GetPageMode", null);
	}

	getViewerPreferences() {
		return this.messageHandler.sendWithPromise("GetViewerPreferences", null);
	}

	getOpenAction() {
		return this.messageHandler.sendWithPromise("GetOpenAction", null);
	}

	getAttachments() {
		return this.messageHandler.sendWithPromise("GetAttachments", null);
	}

	getJavaScript() {
		return this.messageHandler.sendWithPromise("GetJavaScript", null);
	}

	getDocJSActions() {
		return this.messageHandler.sendWithPromise("GetDocJSActions", null);
	}

	getPageJSActions(pageIndex) {
		return this.messageHandler.sendWithPromise("GetPageJSActions", { pageIndex });
	}

	getStructTree(pageIndex) {
		return this.messageHandler.sendWithPromise("GetStructTree", { pageIndex });
	}

	getOutline() {
		return this.messageHandler.sendWithPromise("GetOutline", null);
	}

	getOptionalContentConfig() {
		return this.messageHandler.sendWithPromise("GetOptionalContentConfig", null).then(results => {
			return new OptionalContentConfig(results);
		});
	}

	getPermissions() {
		return this.messageHandler.sendWithPromise("GetPermissions", null);
	}

	getMetadata() {
		const name = "GetMetadata", cachedPromise = this.#methodPromises.get(name);

		if (cachedPromise) {
			return cachedPromise;
		}

		const promise = this.messageHandler.sendWithPromise(name, null).then(results => {
			return {
				info: results[0],
				metadata: results[1] ? new Metadata(results[1]) : null,
				contentDispositionFilename: this._fullReader?.filename ?? null,
				contentLength: this._fullReader?.contentLength ?? null
			};
		});

		this.#methodPromises.set(name, promise);
		return promise;
	}

	getMarkInfo() {
		return this.messageHandler.sendWithPromise("GetMarkInfo", null);
	}

	async startCleanup(keepLoadedFonts = false) {
		if (this.destroyed) {
			return;
		}

		await this.messageHandler.sendWithPromise("Cleanup", null);

		for (const page of this.#pageCache.values()) {
			const cleanupSuccessful = page.cleanup();

			if (!cleanupSuccessful) {
				throw new Error(`startCleanup: Page ${page.pageNumber} is currently rendering.`);
			}
		}

		this.commonObjs.clear();

		if (!keepLoadedFonts) {
			this.fontLoader.clear();
		}

		this.#methodPromises.clear();
		this.filterFactory.destroy(true);
	}

	get loadingParams() {
		const { disableAutoFetch, enableXfa } = this._params;
		return shadow(this, "loadingParams", { disableAutoFetch, enableXfa });
	}
}

class PDFObjects {
	#objs = Object.create(null);

	#ensureObj(objId) {
		const obj = this.#objs[objId];

		if (obj) {
			return obj;
		}

		return this.#objs[objId] = {
			capability: createPromiseCapability(),
			data: null
		};
	}

	get(objId, callback = null) {
		if (callback) {
			const obj = this.#ensureObj(objId);
			obj.capability.promise.then(() => callback(obj.data));
			return null;
		}

		const obj = this.#objs[objId];

		if (!obj?.capability.settled) {
			throw new Error(`Requesting object that isn't resolved yet ${objId}.`);
		}

		return obj.data;
	}

	has(objId) {
		const obj = this.#objs[objId];
		return obj?.capability.settled || false;
	}

	resolve(objId, data = null) {
		const obj = this.#ensureObj(objId);
		obj.data = data;
		obj.capability.resolve();
	}

	clear() {
		for (const objId in this.#objs) {
			const { data } = this.#objs[objId];
			data?.bitmap?.close();
		}

		this.#objs = Object.create(null);
	}
}

class RenderTask {
	#internalRenderTask = null;

	constructor(internalRenderTask) {
		this.#internalRenderTask = internalRenderTask;
		this.onContinue = null;
	}

	get promise() {
		return this.#internalRenderTask.capability.promise;
	}

	cancel(extraDelay = 0) {
		this.#internalRenderTask.cancel(null, extraDelay);
	}

	get separateAnnots() {
		const { separateAnnots } = this.#internalRenderTask.operatorList;

		if (!separateAnnots) {
			return false;
		}

		const { annotationCanvasMap } = this.#internalRenderTask;
		return separateAnnots.form || separateAnnots.canvas && annotationCanvasMap?.size > 0;
	}
}

class InternalRenderTask {
	static #canvasInUse = new WeakSet();

	constructor({ callback, params, objs, commonObjs, annotationCanvasMap, operatorList, pageIndex, canvasFactory, filterFactory, useRequestAnimationFrame = false, pdfBug = false, pageColors = null }) {
		this.callback = callback;
		this.params = params;
		this.objs = objs;
		this.commonObjs = commonObjs;
		this.annotationCanvasMap = annotationCanvasMap;
		this.operatorListIdx = null;
		this.operatorList = operatorList;
		this._pageIndex = pageIndex;
		this.canvasFactory = canvasFactory;
		this.filterFactory = filterFactory;
		this._pdfBug = pdfBug;
		this.pageColors = pageColors;
		this.running = false;
		this.graphicsReadyCallback = null;
		this.graphicsReady = false;
		this._useRequestAnimationFrame = useRequestAnimationFrame === true && typeof window !== "undefined";
		this.cancelled = false;
		this.capability = createPromiseCapability();
		this.task = new RenderTask(this);
		this._cancelBound = this.cancel.bind(this);
		this._continueBound = this._continue.bind(this);
		this._scheduleNextBound = this._scheduleNext.bind(this);
		this._nextBound = this._next.bind(this);
		this._canvas = params.canvasContext.canvas;
	}

	get completed() {
		return this.capability.promise.catch(function () {
			
		});
	}

	initializeGraphics({ transparency = false, optionalContentConfig }) {
		if (this.cancelled) {
			return;
		}

		if (this._canvas) {
			if (InternalRenderTask.#canvasInUse.has(this._canvas)) {
				throw new Error("Cannot use the same canvas during multiple render() operations. " + "Use different canvas or ensure previous operations were " + "cancelled or completed.");
			}

			InternalRenderTask.#canvasInUse.add(this._canvas);
		}

		if (this._pdfBug && globalThis.StepperManager?.enabled) {
			this.stepper = globalThis.StepperManager.create(this._pageIndex);
			this.stepper.init(this.operatorList);
			this.stepper.nextBreakPoint = this.stepper.getNextBreakPoint();
		}

		const { canvasContext, viewport, transform, background } = this.params;
		this.gfx = new CanvasGraphics(canvasContext, this.commonObjs, this.objs, this.canvasFactory, this.filterFactory, { optionalContentConfig }, this.annotationCanvasMap);

		this.gfx.beginDrawing({
			transform,
			viewport,
			transparency,
			background
		});

		this.operatorListIdx = 0;
		this.graphicsReady = true;
		this.graphicsReadyCallback?.();
	}

	cancel(error = null, extraDelay = 0) {
		this.running = false;
		this.cancelled = true;
		this.gfx?.endDrawing(this.pageColors);

		if (this._canvas) {
			InternalRenderTask.#canvasInUse.delete(this._canvas);
		}

		this.callback(error || new RenderingCancelledException(`Rendering cancelled, page ${this._pageIndex + 1}`, "canvas", extraDelay));
	}

	operatorListChanged() {
		if (!this.graphicsReady) {
			if (!this.graphicsReadyCallback) {
				this.graphicsReadyCallback = this._continueBound;
			}

			return;
		}

		this.stepper?.updateOperatorList(this.operatorList);

		if (this.running) {
			return;
		}

		this._continue();
	}

	_continue() {
		this.running = true;

		if (this.cancelled) {
			return;
		}

		if (this.task.onContinue) {
			this.task.onContinue(this._scheduleNextBound);
		} else {
			this._scheduleNext();
		}
	}

	_scheduleNext() {
		if (this._useRequestAnimationFrame) {
			window.requestAnimationFrame(() => {
				this._nextBound().catch(this._cancelBound);
			});
		} else {
			Promise.resolve().then(this._nextBound).catch(this._cancelBound);
		}
	}

	async _next() {
		if (this.cancelled) {
			return;
		}

		this.operatorListIdx = this.gfx.executeOperatorList(this.operatorList, this.operatorListIdx, this._continueBound, this.stepper);

		if (this.operatorListIdx === this.operatorList.argsArray.length) {
			this.running = false;

			if (this.operatorList.lastChunk) {
				this.gfx.endDrawing(this.pageColors);

				if (this._canvas) {
					InternalRenderTask.#canvasInUse.delete(this._canvas);
				}

				this.callback();
			}
		}
	}
}

const version = "ciri-custom";
const build = "ciri-custom";

export { DefaultCMapReaderFactory, DefaultCanvasFactory, DefaultFilterFactory, DefaultStandardFontDataFactory, LoopbackPort, PDFDataRangeTransport, PDFDocumentLoadingTask, PDFDocumentProxy, PDFPageProxy, PDFWorker, PDFWorkerUtil, RenderTask, build, getDocument, version };
