import { BaseCMapReaderFactory, BaseStandardFontDataFactory, BaseSVGFactory, BaseFilterFactory, BaseCanvasFactory } from './base_factory.js';
import { BaseException, warn, shadow, Util, stringToBytes } from '../shared/util.js';

const SVG_NS = "http://www.w3.org/2000/svg";
const AnnotationPrefix = "pdfjs_internal_id_";

class PixelsPerInch {
	static CSS = 96.0;
	static PDF = 72.0;
	static PDF_TO_CSS_UNITS = this.CSS / this.PDF;
}

class DOMFilterFactory extends BaseFilterFactory {
	#_cache;
	#_defs;
	#docId;
	#document;
	#hcmFilter;
	#hcmKey;
	#hcmUrl;
	#id = 0;

	constructor({ docId, ownerDocument = globalThis.document } = {}) {
		super();
		this.#docId = docId;
		this.#document = ownerDocument;
	}

	get #cache() {
		return this.#_cache ||= new Map();
	}

	get #defs() {
		if (!this.#_defs) {
			const div = this.#document.createElement("div");
			const { style } = div;
			style.visibility = "hidden";
			style.contain = "strict";
			style.width = style.height = 0;
			style.position = "absolute";
			style.top = style.left = 0;
			style.zIndex = -1;
			const svg = this.#document.createElementNS(SVG_NS, "svg");
			svg.setAttribute("width", 0);
			svg.setAttribute("height", 0);
			this.#_defs = this.#document.createElementNS(SVG_NS, "defs");
			div.append(svg);
			svg.append(this.#_defs);
			this.#document.body.append(div);
		}

		return this.#_defs;
	}

	#appendFeFunc(feComponentTransfer, func, table) {
		const feFunc = this.#document.createElementNS(SVG_NS, func);
		feFunc.setAttribute("type", "discrete");
		feFunc.setAttribute("tableValues", table);
		feComponentTransfer.append(feFunc);
	}

	addFilter(maps) {
		if (!maps) {
			return "none";
		}

		let value = this.#cache.get(maps);

		if (value) {
			return value;
		}

		let tableR, tableG, tableB, key;

		if (maps.length === 1) {
			const mapR = maps[0];
			const buffer = new Array(256);

			for (let i = 0; i < 256; i++) {
				buffer[i] = mapR[i] / 255;
			}

			key = tableR = tableG = tableB = buffer.join(",");
		} else {
			const [mapR, mapG, mapB] = maps;
			const bufferR = new Array(256);
			const bufferG = new Array(256);
			const bufferB = new Array(256);

			for (let i = 0; i < 256; i++) {
				bufferR[i] = mapR[i] / 255;
				bufferG[i] = mapG[i] / 255;
				bufferB[i] = mapB[i] / 255;
			}

			tableR = bufferR.join(",");
			tableG = bufferG.join(",");
			tableB = bufferB.join(",");
			key = `${tableR}${tableG}${tableB}`;
		}

		value = this.#cache.get(key);

		if (value) {
			this.#cache.set(maps, value);
			return value;
		}

		const id = `g_${this.#docId}_transfer_map_${this.#id++}`;
		const url = `url(#${id})`;
		this.#cache.set(maps, url);
		this.#cache.set(key, url);
		const filter = this.#document.createElementNS(SVG_NS, "filter", SVG_NS);
		filter.setAttribute("id", id);
		filter.setAttribute("color-interpolation-filters", "sRGB");
		const feComponentTransfer = this.#document.createElementNS(SVG_NS, "feComponentTransfer");
		filter.append(feComponentTransfer);
		this.#appendFeFunc(feComponentTransfer, "feFuncR", tableR);
		this.#appendFeFunc(feComponentTransfer, "feFuncG", tableG);
		this.#appendFeFunc(feComponentTransfer, "feFuncB", tableB);
		this.#defs.append(filter);
		return url;
	}

	addHCMFilter(fgColor, bgColor) {
		const key = `${fgColor}-${bgColor}`;

		if (this.#hcmKey === key) {
			return this.#hcmUrl;
		}

		this.#hcmKey = key;
		this.#hcmUrl = "none";
		this.#hcmFilter?.remove();

		if (!fgColor || !bgColor) {
			return this.#hcmUrl;
		}

		this.#defs.style.color = fgColor;
		fgColor = getComputedStyle(this.#defs).getPropertyValue("color");
		const fgRGB = getRGB(fgColor);
		fgColor = Util.makeHexColor(...fgRGB);
		this.#defs.style.color = bgColor;
		bgColor = getComputedStyle(this.#defs).getPropertyValue("color");
		const bgRGB = getRGB(bgColor);
		bgColor = Util.makeHexColor(...bgRGB);
		this.#defs.style.color = "";

		if (fgColor === "#000000" && bgColor === "#ffffff" || fgColor === bgColor) {
			return this.#hcmUrl;
		}

		const map = new Array(256);

		for (let i = 0; i <= 255; i++) {
			const x = i / 255;
			map[i] = x <= 0.03928 ? x / 12.92 : ((x + 0.055) / 1.055) ** 2.4;
		}

		const table = map.join(",");
		const id = `g_${this.#docId}_hcm_filter`;
		const filter = this.#hcmFilter = this.#document.createElementNS(SVG_NS, "filter", SVG_NS);
		filter.setAttribute("id", id);
		filter.setAttribute("color-interpolation-filters", "sRGB");
		let feComponentTransfer = this.#document.createElementNS(SVG_NS, "feComponentTransfer");
		filter.append(feComponentTransfer);
		this.#appendFeFunc(feComponentTransfer, "feFuncR", table);
		this.#appendFeFunc(feComponentTransfer, "feFuncG", table);
		this.#appendFeFunc(feComponentTransfer, "feFuncB", table);
		const feColorMatrix = this.#document.createElementNS(SVG_NS, "feColorMatrix");
		feColorMatrix.setAttribute("type", "matrix");
		feColorMatrix.setAttribute("values", "0.2126 0.7152 0.0722 0 0 0.2126 0.7152 0.0722 0 0 0.2126 0.7152 0.0722 0 0 0 0 0 1 0");
		filter.append(feColorMatrix);
		feComponentTransfer = this.#document.createElementNS(SVG_NS, "feComponentTransfer");
		filter.append(feComponentTransfer);

		const getSteps = (c, n) => {
			const start = fgRGB[c] / 255;
			const end = bgRGB[c] / 255;
			const arr = new Array(n + 1);

			for (let i = 0; i <= n; i++) {
				arr[i] = start + i / n * (end - start);
			}

			return arr.join(",");
		};

		this.#appendFeFunc(feComponentTransfer, "feFuncR", getSteps(0, 5));
		this.#appendFeFunc(feComponentTransfer, "feFuncG", getSteps(1, 5));
		this.#appendFeFunc(feComponentTransfer, "feFuncB", getSteps(2, 5));
		this.#defs.append(filter);
		this.#hcmUrl = `url(#${id})`;
		return this.#hcmUrl;
	}

	destroy(keepHCM = false) {
		if (keepHCM && this.#hcmUrl) {
			return;
		}

		if (this.#_defs) {
			this.#_defs.parentNode.parentNode.remove();
			this.#_defs = null;
		}

		if (this.#_cache) {
			this.#_cache.clear();
			this.#_cache = null;
		}

		this.#id = 0;
	}
}

class DOMCanvasFactory extends BaseCanvasFactory {
	constructor({ ownerDocument = globalThis.document } = {}) {
		super();
		this._document = ownerDocument;
	}

	_createCanvas(width, height) {
		const canvas = this._document.createElement("canvas");
		canvas.width = width;
		canvas.height = height;
		return canvas;
	}
}

async function fetchData(url, asTypedArray = false) {
	if (isValidFetchUrl(url, document.baseURI)) {
		const response = await fetch(url);

		if (!response.ok) {
			throw new Error(response.statusText);
		}

		return asTypedArray
		? new Uint8Array(await response.arrayBuffer())
		: stringToBytes(await response.text());
	}

	return new Promise((resolve, reject) => {
			const request = new XMLHttpRequest();
			request.open("GET", url, true);

			if (asTypedArray) {
				request.responseType = "arraybuffer";
			}

			request.onreadystatechange = () => {
				if (request.readyState !== XMLHttpRequest.DONE) {
					return;
				}

				if (request.status === 200 || request.status === 0) {
					let data;

					if (asTypedArray && request.response) {
						data = new Uint8Array(request.response);
					} else if (!asTypedArray && request.responseText) {
						data = stringToBytes(request.responseText);
					}

					if (data) {
						resolve(data);
						return;
					}
				}

				reject(new Error(request.statusText));
			};

			request.send(null);
		});
}

class DOMCMapReaderFactory extends BaseCMapReaderFactory {
	_fetchData(url, compressionType) {
		return fetchData(url, this.isCompressed).then(data => {
			return { cMapData: data, compressionType };
		});
	}
}

class DOMStandardFontDataFactory extends BaseStandardFontDataFactory {
	_fetchData(url) {
		return fetchData(url, true);
	}
}

class DOMSVGFactory extends BaseSVGFactory {
	_createSVG(type) {
		return document.createElementNS(SVG_NS, type);
	}
}

class PageViewport {
	constructor({ viewBox, scale, rotation, offsetX = 0, offsetY = 0, dontFlip = false }) {
		this.viewBox = viewBox;
		this.scale = scale;
		this.rotation = rotation;
		this.offsetX = offsetX;
		this.offsetY = offsetY;
		const centerX = (viewBox[2] + viewBox[0]) / 2;
		const centerY = (viewBox[3] + viewBox[1]) / 2;
		let rotateA, rotateB, rotateC, rotateD;
		rotation %= 360;

		if (rotation < 0) {
			rotation += 360;
		}

		switch (rotation) {
			case 180:
				rotateA = -1;
				rotateB = 0;
				rotateC = 0;
				rotateD = 1;
				break;
			case 90:
				rotateA = 0;
				rotateB = 1;
				rotateC = 1;
				rotateD = 0;
				break;
			case 270:
				rotateA = 0;
				rotateB = -1;
				rotateC = -1;
				rotateD = 0;
				break;
			case 0:
				rotateA = 1;
				rotateB = 0;
				rotateC = 0;
				rotateD = -1;
				break;
			default:
				throw new Error("PageViewport: Invalid rotation, must be a multiple of 90 degrees.");
		}

		if (dontFlip) {
			rotateC = -rotateC;
			rotateD = -rotateD;
		}

		let offsetCanvasX, offsetCanvasY;
		let width, height;

		if (rotateA === 0) {
			offsetCanvasX = Math.abs(centerY - viewBox[1]) * scale + offsetX;
			offsetCanvasY = Math.abs(centerX - viewBox[0]) * scale + offsetY;
			width = (viewBox[3] - viewBox[1]) * scale;
			height = (viewBox[2] - viewBox[0]) * scale;
		} else {
			offsetCanvasX = Math.abs(centerX - viewBox[0]) * scale + offsetX;
			offsetCanvasY = Math.abs(centerY - viewBox[1]) * scale + offsetY;
			width = (viewBox[2] - viewBox[0]) * scale;
			height = (viewBox[3] - viewBox[1]) * scale;
		}

		this.transform = [
			rotateA * scale,
			rotateB * scale,
			rotateC * scale,
			rotateD * scale,
			offsetCanvasX - rotateA * scale * centerX - rotateC * scale * centerY,
			offsetCanvasY - rotateB * scale * centerX - rotateD * scale * centerY
		];

		this.width = width;
		this.height = height;
	}

	get rawDims() {
		const { viewBox } = this;

		return shadow(this, "rawDims", {
			pageWidth: viewBox[2] - viewBox[0],
			pageHeight: viewBox[3] - viewBox[1],
			pageX: viewBox[0],
			pageY: viewBox[1]
		});
	}

	clone({ scale = this.scale, rotation = this.rotation, offsetX = this.offsetX, offsetY = this.offsetY, dontFlip = false } = {}) {
		return new PageViewport({
				viewBox: this.viewBox.slice(),
				scale,
				rotation,
				offsetX,
				offsetY,
				dontFlip
			});
	}

	convertToViewportPoint(x, y) {
		return Util.applyTransform([x, y], this.transform);
	}

	convertToViewportRectangle(rect) {
		const topLeft = Util.applyTransform([rect[0], rect[1]], this.transform);
		const bottomRight = Util.applyTransform([rect[2], rect[3]], this.transform);
		return [topLeft[0], topLeft[1], bottomRight[0], bottomRight[1]];
	}

	convertToPdfPoint(x, y) {
		return Util.applyInverseTransform([x, y], this.transform);
	}
}

class RenderingCancelledException extends BaseException {
	constructor(msg, type, extraDelay = 0) {
		super(msg, "RenderingCancelledException");
		this.type = type;
		this.extraDelay = extraDelay;
	}
}

function isDataScheme(url) {
	const ii = url.length;
	let i = 0;

	while (i < ii && url[i].trim() === "") {
		i++;
	}

	return url.substring(i, i + 5).toLowerCase() === "data:";
}

function isPdfFile(filename) {
	return typeof filename === "string" && (/\.pdf$/i).test(filename);
}

function getFilenameFromUrl(url, onlyStripPath = false) {
	if (!onlyStripPath) {
		[url] = url.split(/[#?]/, 1);
	}

	return url.substring(url.lastIndexOf("/") + 1);
}

function getPdfFilenameFromUrl(url, defaultFilename = "document.pdf") {
	if (typeof url !== "string") {
		return defaultFilename;
	}

	if (isDataScheme(url)) {
		warn('getPdfFilenameFromUrl: ignore "data:"-URL for performance reasons.');
		return defaultFilename;
	}

	const reURI = /^(?:(?:[^:]+:)?\/\/[^/]+)?([^?#]*)(\?[^#]*)?(#.*)?$/;
	const reFilename = /[^/?#=]+\.pdf\b(?!.*\.pdf\b)/i;
	const splitURI = reURI.exec(url);
	let suggestedFilename = reFilename.exec(splitURI[1]) || reFilename.exec(splitURI[2]) || reFilename.exec(splitURI[3]);

	if (suggestedFilename) {
		suggestedFilename = suggestedFilename[0];

		if (suggestedFilename.includes("%")) {
			try {
				suggestedFilename = reFilename.exec(decodeURIComponent(suggestedFilename))[0];
			} catch(ex) {
				
			}
		}
	}

	return suggestedFilename || defaultFilename;
}

class StatTimer {
	started = Object.create(null);
	times = [];

	time(name) {
		if (name in this.started) {
			warn(`Timer is already running for ${name}`);
		}

		this.started[name] = Date.now();
	}

	timeEnd(name) {
		if (!(name in this.started)) {
			warn(`Timer has not been started for ${name}`);
		}

		this.times.push({
			name,
			start: this.started[name],
			end: Date.now()
		});

		delete this.started[name];
	}

	toString() {
		const outBuf = [];
		let longest = 0;

		for (const { name } of this.times) {
			longest = Math.max(name.length, longest);
		}

		for (const { name, start, end } of this.times) {
			outBuf.push(`${name.padEnd(longest)} ${end - start}ms\n`);
		}

		return outBuf.join("");
	}
}

function isValidFetchUrl(url, baseUrl) {
	try {
		const { protocol } = baseUrl ? new URL(url, baseUrl) : new URL(url);
		return protocol === "http:" || protocol === "https:";
	} catch(ex) {
		return false;
	}
}

function loadScript(src, removeScriptElement = false) {
	return new Promise((resolve, reject) => {
			const script = document.createElement("script");
			script.src = src;

			script.onload = function (evt) {
				if (removeScriptElement) {
					script.remove();
				}

				resolve(evt);
			};

			script.onerror = function () {
				reject(new Error(`Cannot load script at: ${script.src}`));
			};

			(document.head || document.documentElement).append(script);
		});
}

function deprecated(details) {
	console.log("Deprecated API usage: " + details);
}

let pdfDateStringRegex;

class PDFDateString {
	static toDateObject(input) {
		if (!input || typeof input !== "string") {
			return null;
		}

		if (!pdfDateStringRegex) {
			pdfDateStringRegex = new RegExp("^D:" + "(\\d{4})" + "(\\d{2})?" + "(\\d{2})?" + "(\\d{2})?" + "(\\d{2})?" + "(\\d{2})?" + "([Z|+|-])?" + "(\\d{2})?" + "'?" + "(\\d{2})?" + "'?");
		}

		const matches = pdfDateStringRegex.exec(input);

		if (!matches) {
			return null;
		}

		const year = parseInt(matches[1], 10);
		let month = parseInt(matches[2], 10);
		month = month >= 1 && month <= 12 ? month - 1 : 0;
		let day = parseInt(matches[3], 10);
		day = day >= 1 && day <= 31 ? day : 1;
		let hour = parseInt(matches[4], 10);
		hour = hour >= 0 && hour <= 23 ? hour : 0;
		let minute = parseInt(matches[5], 10);
		minute = minute >= 0 && minute <= 59 ? minute : 0;
		let second = parseInt(matches[6], 10);
		second = second >= 0 && second <= 59 ? second : 0;
		const universalTimeRelation = matches[7] || "Z";
		let offsetHour = parseInt(matches[8], 10);
		offsetHour = offsetHour >= 0 && offsetHour <= 23 ? offsetHour : 0;
		let offsetMinute = parseInt(matches[9], 10) || 0;

		offsetMinute = offsetMinute >= 0 && offsetMinute <= 59
		? offsetMinute
		: 0;

		if (universalTimeRelation === "-") {
			hour += offsetHour;
			minute += offsetMinute;
		} else if (universalTimeRelation === "+") {
			hour -= offsetHour;
			minute -= offsetMinute;
		}

		return new Date(Date.UTC(year, month, day, hour, minute, second));
	}
}

function getXfaPageViewport(xfaPage, { scale = 1, rotation = 0 }) {
	const { width, height } = xfaPage.attributes.style;
	const viewBox = [0, 0, parseInt(width), parseInt(height)];
	return new PageViewport({ viewBox, scale, rotation });
}

function getRGB(color) {
	if (color.startsWith("#")) {
		const colorRGB = parseInt(color.slice(1), 16);
		return [(colorRGB & 0xff0000) >> 16, (colorRGB & 0x00ff00) >> 8, colorRGB & 0x0000ff];
	}

	if (color.startsWith("rgb(")) {
		return color.slice(4, -1).split(",").map(x => parseInt(x));
	}

	if (color.startsWith("rgba(")) {
		return color.slice(5, -1).split(",").map(x => parseInt(x)).slice(0, 3);
	}

	warn(`Not a valid color format: "${color}"`);
	return [0, 0, 0];
}

function getColorValues(colors) {
	const span = document.createElement("span");
	span.style.visibility = "hidden";
	document.body.append(span);

	for (const name of colors.keys()) {
		span.style.color = name;
		const computedColor = window.getComputedStyle(span).color;
		colors.set(name, getRGB(computedColor));
	}

	span.remove();
}

function getCurrentTransform(ctx) {
	const { a, b, c, d, e, f } = ctx.getTransform();
	return [a, b, c, d, e, f];
}

function getCurrentTransformInverse(ctx) {
	const { a, b, c, d, e, f } = ctx.getTransform().invertSelf();
	return [a, b, c, d, e, f];
}

function setLayerDimensions(div, viewport, mustFlip = false, mustRotate = true) {
	if (viewport instanceof PageViewport) {
		const { pageWidth, pageHeight } = viewport.rawDims;
		const { style } = div;
		const widthStr = `calc(var(--scale-factor) * ${pageWidth}px)`;
		const heightStr = `calc(var(--scale-factor) * ${pageHeight}px)`;

		if (!mustFlip || viewport.rotation % 180 === 0) {
			style.width = widthStr;
			style.height = heightStr;
		} else {
			style.width = heightStr;
			style.height = widthStr;
		}
	}

	if (mustRotate) {
		div.setAttribute("data-main-rotation", viewport.rotation);
	}
}

export { AnnotationPrefix, DOMCMapReaderFactory, DOMCanvasFactory, DOMFilterFactory, DOMSVGFactory, DOMStandardFontDataFactory, PDFDateString, PageViewport, PixelsPerInch, RenderingCancelledException, StatTimer, deprecated, getColorValues, getCurrentTransform, getCurrentTransformInverse, getFilenameFromUrl, getPdfFilenameFromUrl, getRGB, getXfaPageViewport, isDataScheme, isPdfFile, isValidFetchUrl, loadScript, setLayerDimensions };
