import { XfaText } from './xfa_text.js';

class XfaLayer {
	static setupStorage(html, id, element, storage, intent) {
		const storedData = storage.getValue(id, { value: null });

		switch (element.name) {
			case "textarea":
				if (storedData.value !== null) {
					html.textContent = storedData.value;
				}
				if (intent === "print") {
					break;
				}
				html.addEventListener("input", event => {
					storage.setValue(id, { value: event.target.value });
				});
				break;
			case "input":
				if (element.attributes.type === "radio" || element.attributes.type === "checkbox") {
					if (storedData.value === element.attributes.xfaOn) {
						html.setAttribute("checked", true);
					} else if (storedData.value === element.attributes.xfaOff) {
						html.removeAttribute("checked");
					}

					if (intent === "print") {
						break;
					}

					html.addEventListener("change", event => {
						storage.setValue(id, {
							value: event.target.checked
							? event.target.getAttribute("xfaOn")
							: event.target.getAttribute("xfaOff")
						});
					});
				} else {
					if (storedData.value !== null) {
						html.setAttribute("value", storedData.value);
					}

					if (intent === "print") {
						break;
					}

					html.addEventListener("input", event => {
						storage.setValue(id, { value: event.target.value });
					});
				}
				break;
			case "select":
				if (storedData.value !== null) {
					for (const option of element.children) {
						if (option.attributes.value === storedData.value) {
							option.attributes.selected = true;
						}
					}
				}
				html.addEventListener("input", event => {
					const options = event.target.options;

					const value = options.selectedIndex === -1
					? ""
					: options[options.selectedIndex].value;

					storage.setValue(id, { value });
				});
				break;
		}
	}

	static setAttributes({ html, element, storage = null, intent, linkService }) {
		const { attributes } = element;
		const isHTMLAnchorElement = html instanceof HTMLAnchorElement;

		if (attributes.type === "radio") {
			attributes.name = `${attributes.name}-${intent}`;
		}

		for (const [key, value] of Object.entries(attributes)) {
			if (value === null || value === undefined) {
				continue;
			}

			switch (key) {
				case "class":
					if (value.length) {
						html.setAttribute(key, value.join(" "));
					}
					break;
				case "dataId":
					break;
				case "id":
					html.setAttribute("data-element-id", value);
					break;
				case "style":
					Object.assign(html.style, value);
					break;
				case "textContent":
					html.textContent = value;
					break;
				default:
					if (!isHTMLAnchorElement || key !== "href" && key !== "newWindow") {
						html.setAttribute(key, value);
					}
			}
		}

		if (isHTMLAnchorElement) {
			linkService.addLinkAttributes(html, attributes.href, attributes.newWindow);
		}

		if (storage && attributes.dataId) {
			this.setupStorage(html, attributes.dataId, element, storage);
		}
	}

	static render(parameters) {
		const storage = parameters.annotationStorage;
		const linkService = parameters.linkService;
		const root = parameters.xfaHtml;
		const intent = parameters.intent || "display";
		const rootHtml = document.createElement(root.name);

		if (root.attributes) {
			this.setAttributes({
				html: rootHtml,
				element: root,
				intent,
				linkService
			});
		}

		const stack = [[root, -1, rootHtml]];
		const rootDiv = parameters.div;
		rootDiv.append(rootHtml);

		if (parameters.viewport) {
			const transform = `matrix(${parameters.viewport.transform.join(",")})`;
			rootDiv.style.transform = transform;
		}

		if (intent !== "richText") {
			rootDiv.setAttribute("class", "xfaLayer xfaFont");
		}

		const textDivs = [];

		while (stack.length > 0) {
			const [parent, i, html] = stack.at(-1);

			if (i + 1 === parent.children.length) {
				stack.pop();
				continue;
			}

			const child = parent.children[++stack.at(-1)[1]];

			if (child === null) {
				continue;
			}

			const { name } = child;

			if (name === "#text") {
				const node = document.createTextNode(child.value);
				textDivs.push(node);
				html.append(node);
				continue;
			}

			let childHtml;

			if (child?.attributes?.xmlns) {
				childHtml = document.createElementNS(child.attributes.xmlns, name);
			} else {
				childHtml = document.createElement(name);
			}

			html.append(childHtml);

			if (child.attributes) {
				this.setAttributes({
					html: childHtml,
					element: child,
					storage,
					intent,
					linkService
				});
			}

			if (child.children && child.children.length > 0) {
				stack.push([child, -1, childHtml]);
			} else if (child.value) {
				const node = document.createTextNode(child.value);

				if (XfaText.shouldBuildText(name)) {
					textDivs.push(node);
				}

				childHtml.append(node);
			}
		}

		for (const el of rootDiv.querySelectorAll(".xfaNonInteractive input, .xfaNonInteractive textarea")) {
			el.setAttribute("readOnly", true);
		}

		return { textDivs };
	}

	static update(parameters) {
		const transform = `matrix(${parameters.viewport.transform.join(",")})`;
		parameters.div.style.transform = transform;
		parameters.div.hidden = false;
	}
}

export { XfaLayer };
